Package: Dino
Type: Package
Title: Normalization of Single-Cell mRNA Sequencing Data
Version: 1.18.0
Authors@R: c(
    person("Jared", "Brown", email = "brownj@ds.dfci.harvard.edu",
    role = c("aut", "cre"), comment = c(ORCID = "0000-0002-9151-4386")),
    person("Christina", "Kendziorski", role= c("ctb")))
biocViews: Software, Normalization, RNASeq, SingleCell, Sequencing,
        GeneExpression, Transcriptomics, Regression, CellBasedAssays
Description: Dino normalizes single-cell, mRNA sequencing data to
        correct for technical variation, particularly sequencing depth,
        prior to downstream analysis. The approach produces a matrix of
        corrected expression for which the dependency between
        sequencing depth and the full distribution of normalized
        expression; many existing methods aim to remove only the
        dependency between sequencing depth and the mean of the
        normalized expression. This is particuarly useful in the
        context of highly sparse datasets such as those produced by 10X
        genomics and other uninque molecular identifier (UMI) based
        microfluidics protocols for which the depth-dependent
        proportion of zeros in the raw expression data can otherwise
        present a challenge.
Depends: R (>= 4.0.0)
License: GPL-3
Encoding: UTF-8
LazyData: false
RoxygenNote: 7.1.1
Suggests: testthat (>= 2.1.0), knitr, rmarkdown, BiocStyle, devtools,
        ggplot2, gridExtra, ggpubr, grid, magick, hexbin
VignetteBuilder: knitr
Imports: BiocParallel, BiocSingular, SummarizedExperiment,
        SingleCellExperiment, S4Vectors, Matrix, Seurat, matrixStats,
        parallel, scran, grDevices, stats, methods
URL: https://github.com/JBrownBiostat/Dino
BugReports: https://github.com/JBrownBiostat/Dino/issues
Config/pak/sysreqs: cmake libglpk-dev make libicu-dev libpng-dev
        libuv1-dev libxml2-dev libssl-dev python3 zlib1g-dev
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 12:56:29 UTC
RemoteUrl: https://github.com/bioc/Dino
RemoteRef: RELEASE_3_23
RemoteSha: a4aa74c9092bacb6dac689f888ae3669cbd73b0e
NeedsCompilation: no
Packaged: 2026-04-29 21:01:53 UTC; root
Author: Jared Brown [aut, cre] (ORCID: <https://orcid.org/0000-0002-9151-4386>),
  Christina Kendziorski [ctb]
Maintainer: Jared Brown <brownj@ds.dfci.harvard.edu>
Built: R 4.6.0; ; 2026-04-29 21:46:20 UTC; unix

# Generated by roxygen2: do not edit by hand

export(Dino)
export(Dino_SCE)
export(SeuratFromDino)
import(SingleCellExperiment)
importClassesFrom(Matrix,dgCMatrix)
importFrom(BiocParallel,MulticoreParam)
importFrom(BiocParallel,SnowParam)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,register)
importFrom(BiocSingular,IrlbaParam)
importFrom(Matrix,Matrix)
importFrom(Matrix,colSums)
importFrom(Matrix,rowSums)
importFrom(Matrix,t)
importFrom(S4Vectors,metadata)
importFrom(Seurat,CreateAssayObject)
importFrom(Seurat,CreateSeuratObject)
importFrom(Seurat,Misc)
importFrom(SummarizedExperiment,assayNames)
importFrom(grDevices,chull)
importFrom(matrixStats,colMaxs)
importFrom(matrixStats,rowMaxs)
importFrom(methods,is)
importFrom(methods,new)
importFrom(parallel,detectCores)
importFrom(scran,quickCluster)
importFrom(stats,approxfun)
importFrom(stats,bw.nrd)
importFrom(stats,bw.ucv)
importFrom(stats,density)
importFrom(stats,glm)
importFrom(stats,median)
importFrom(stats,optimize)
importFrom(stats,pgamma)
importFrom(stats,ppois)
importFrom(stats,qgamma)
importFrom(stats,rgamma)
importFrom(stats,runif)

# Dino
## Dino is an R package for the normalization of single-cell RNA-seq data using a flexible mixture of negative binomials model of expression

This package was developed by *Jared Brown* in *Christina Kendziorski's* lab at the University of Wisconsin-Madison.

Normalization to remove technical or experimental artifacts is critical in the analysis of single-cell RNA-sequencing experiments, even those for which unique molecular identifiers (UMIs) are available. The majority of methods for normalizing single-cell RNA-sequencing data adjust average expression in sequencing depth, but allow the variance and other properties of the gene-specific expression distribution to be non-constant in depth, which often results in reduced power and increased false discoveries in downstream analyses. This problem is exacerbated by the high proportion of zeros present in most datasets.

To address this, *Dino* constructs a flexible negative-binomial mixture model of gene expression. The data are then normalized by sampling from the posterior distribution of expected expression conditional on observed sequencing depth.

## Installation

`Dino` is now available on `BioConductor` and can be easily installed from that repository by running:

```
# Install Bioconductor if not present, skip otherwise
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

# Install Dino package
BiocManager::install("Dino")

# View vignette from R
browseVignettes("Dino")
```

`Dino` is also available from Github, and bug fixes, patches, and updates are available there first. To install `Dino` from Github, run

```
devtools::install_github('JBrownBiostat/Dino', build_vignettes = TRUE)
```

*Note:* building the vignette can take a few minutes. If you do not require the vignette, consider running with `build_vignettes = FALSE` to save time.

## Vignette

In addition to the option to view the package vignette from `R` (see above), a compiled vinette is also available from the `Dino` page on BioConductor: [http://www.bioconductor.org/packages/release/bioc/html/Dino.html](http://www.bioconductor.org/packages/release/bioc/html/Dino.html)

The vignette includes a fuller description of use cases including code examles as well as the underlying methematics of the method.

## Quick start

Following installation, the single funtion, `Dino` can be used to return a normalized matrix of gene expression data:

```
normMat <- Dino(rawMat)
```

For further details on implementation, options, and variations, consult the vignette available by running:

```
vignette("Dino")
```

## Code repository

In addition to [BioConductor](http://www.bioconductor.org/packages/release/bioc/html/Dino.html) and GitHub, *Dino* is further freely available for download from Zenodo:
[https://zenodo.org/record/4897558#.YLjjnW5Okko](https://zenodo.org/record/4897558#.YLjjnW5Okko)

## Citation

If you use `Dino` in your analysis, please cite our paper:

Brown, J., Ni, Z., Mohanty, C., Bacher, R., and Kendziorski, C. (2021). Normalization by distributional resampling of high throughput single-cell RNA-sequencing data. Bioinformatics, 37, 4123-4128. [https://academic.oup.com/bioinformatics/article/37/22/4123/6306403](https://academic.oup.com/bioinformatics/article/37/22/4123/6306403)

## Contact

With questions, comments, or concerns regarding the `Dino` package, please consider opening an issue on Github. You can also contact us directly:

Jared Brown: ![](/vignettes/JBrownEmail.jpg)

Christina Kendziorski: ![](/vignettes/CKendzEmail.jpg)

## ----Initialize, echo=FALSE, results="hide", message=FALSE--------------------
require(knitr)
opts_chunk$set(error=FALSE, message=FALSE, warning=FALSE)

## ----Install Dino BioC, eval = FALSE------------------------------------------
# # Install Bioconductor if not present, skip otherwise
# if (!require("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# # Install Dino package
# BiocManager::install("Dino")
# 
# # View (this) vignette from R
# browseVignettes("Dino")

## ----Install Dino Github, eval = FALSE----------------------------------------
# devtools::install_github('JBrownBiostat/Dino', build_vignettes = TRUE)

## ----Quick Start, eval = FALSE------------------------------------------------
# library(Dino)
# 
# # Return a sparse matrix of normalized expression
# Norm_Mat <- Dino(UMI_Mat)
# 
# # Return a Seurat object from already normalized expression
# # Use normalized (doNorm = FALSE) and un-transformed (doLog = FALSE) expression
# Norm_Seurat <- SeuratFromDino(Norm_Mat, doNorm = FALSE, doLog = FALSE)
# 
# # Return a Seurat object from UMI expression
# # Transform normalized expression as log(x + 1) to improve
# # some types of downstream analysis
# Norm_Seurat <- SeuratFromDino(UMI_Mat)

## ----load pbmcSmall data------------------------------------------------------
set.seed(1)

# Bring pbmcSmall into R environment
library(Dino)
library(Seurat)
library(Matrix)
data("pbmcSmall")
print(dim(pbmcSmall))

## ----clean data---------------------------------------------------------------
# Filter genes for a minimum of non-zero expression
pbmcSmall <- pbmcSmall[rowSums(pbmcSmall != 0) >= 20, ]
print(dim(pbmcSmall))

## ----normalize data, eval = FALSE---------------------------------------------
# # Normalize data
# pbmcSmall_Norm <- Dino(pbmcSmall)

## ----normalize data background, echo = FALSE----------------------------------
invisible(capture.output(pbmcSmall_Norm <- Dino(pbmcSmall)))

## ----Seurat clustering--------------------------------------------------------
# Reformat normalized expression as a Seurat object
pbmcSmall_Seurat <- SeuratFromDino(pbmcSmall_Norm, doNorm = FALSE)

# Cluster pbmcSmall_Seurat
pbmcSmall_Seurat <- FindVariableFeatures(pbmcSmall_Seurat, 
                        selection.method = "mvp")
pbmcSmall_Seurat <- ScaleData(pbmcSmall_Seurat, 
                        features = rownames(pbmcSmall_Norm))
pbmcSmall_Seurat <- RunPCA(pbmcSmall_Seurat, 
                        features = VariableFeatures(object = pbmcSmall_Seurat),
                        verbose = FALSE)
pbmcSmall_Seurat <- FindNeighbors(pbmcSmall_Seurat, dims = 1:10)
pbmcSmall_Seurat <- FindClusters(pbmcSmall_Seurat, verbose = FALSE)
pbmcSmall_Seurat <- RunUMAP(pbmcSmall_Seurat, dims = 1:10)
DimPlot(pbmcSmall_Seurat, reduction = "umap")

## ----SinglleCellExperiment, eval = FALSE--------------------------------------
# # Reformatting pbmcSmall as a SingleCellExperiment
# library(SingleCellExperiment)
# pbmc_SCE <- SingleCellExperiment(assays = list("counts" = pbmcSmall))
# 
# # Run Dino
# pbmc_SCE <- Dino_SCE(pbmc_SCE)
# str(normcounts(pbmc_SCE))

## ----SinglleCellExperiment Background, echo = F-------------------------------
# Reformatting pbmcSmall as a SingleCellExperiment
library(SingleCellExperiment)
pbmc_SCE <- SingleCellExperiment(assays = list("counts" = pbmcSmall))

# Run Dino
invisible(capture.output(pbmc_SCE <- Dino_SCE(pbmc_SCE)))
str(normcounts(pbmc_SCE))

## ----Scran depths, eval = FALSE-----------------------------------------------
# library(scran)
# 
# # Compute scran size factors
# scranSizes <- calculateSumFactors(pbmcSmall)
# 
# # Re-normalize data
# pbmcSmall_SNorm <- Dino(pbmcSmall, nCores = 1, depth = log(scranSizes))

## ----Unimodal Simulation, echo = FALSE, warning = FALSE-----------------------
library(ggplot2)
library(gridExtra)
library(ggpubr)
library(grid)

themeObj <- theme(
    title = element_text(size = 17),
    plot.subtitle = element_text(size = 15),
    axis.title = element_text(size = 14),
    axis.text = element_text(size = 12),
    strip.text = element_text(size = 14),
    legend.title = element_text(size = 14),
    legend.text = element_text(size = 12)
)

p1_func <- function(simDat, dinoLam, muVec, themeObj) {
    p1 <- ggplot(simDat, aes(x = LS, y = y)) +
        theme_classic() +
        geom_hex(aes(fill = log10(..count..))) +
        scale_fill_viridis_c() +
        labs(
            x = "log-Depth", y = "Expression (log)",
            title = "Fitted means", fill = "Count\n(log10)"
        )
    for(i in 1:length(dinoLam)) {
        p1 <- p1 +
            geom_abline(
                slope = 1, intercept = log1p(dinoLam[i]),
                col = 1, alpha = 0.5, lwd = 0.75
            )
    }
    for(i in 1:length(muVec)) {
        p1 <- p1 +
            geom_abline(
                slope = 1, intercept = log1p(muVec[i]),
                col = 2, lwd = 1.5
            )
    }
    p1 <- p1 +
        themeObj
}

p2_func <- function(plotDat, themeObj) {
    x.dens <- plotDat$x[plotDat$model == "NB"]
    y.dens <- plotDat$y[plotDat$model == "NB"]
    p2 <- ggplot(plotDat, aes(x = x, y = y, color = model)) +
        theme_classic() +
        geom_line() +
        scale_color_manual(values = 1:2) +
        labs(
            x = "Expression", y = "Density", color = "Model",
            title = "Reference vs.\nFitted distribution"
        ) +
        themeObj
    xInd <- which.max(
        y.dens < 1e-3 * max(y.dens) &
            x.dens > x.dens[which.max(y.dens)]
    )
    if(xInd > 1) {
        p2 <- p2 +
            xlim(c(
                0, 
                x.dens[which.max(
                    y.dens < 1e-3 * max(y.dens) &
                        x.dens > x.dens[which.max(y.dens)]
                )]
            ))
    }
    dinoDens <- data.frame(
        x = plotDat$x[plotDat$model == "Dino"],
        y = plotDat$y[plotDat$model == "Dino"]
    )
    if(max(y.dens) < max(dinoDens$y)) {
        p2 <- p2 +
            ylim(c(0, min(c(1.025 * max(dinoDens$y), 1.5 * max(y.dens)))))
    }
    return(p2)
}

plotSim_func <- function(plotList, themeObj) {
    p1 <- p1_func(plotList$simDat, plotList$dinoLam, plotList$muVec, themeObj)
    p2 <- p2_func(plotList$plotDat, themeObj)
    
    p <- grid.arrange(
        p1, p2, nrow = 1,
        top = textGrob(paste0("K = ", plotList$k), gp = gpar(fontsize = 17))
    )
    return(p)
}

data("unimodalDat")
plotSim_func(unimodalDat, themeObj)

## ----Multimodal Simulation, echo = FALSE, warning = FALSE---------------------
data("multimodalDat")
plotSim_func(multimodalDat, themeObj)

## -----------------------------------------------------------------------------
sessionInfo()

#' Subset of 200 peripheral blood mononuclear cells (PBMCs) from a healthy donor
#'
#' This dataset derives from the "3k PBMCs from a Healthy Donor" public dataset
#' from 10X Genomics.
#'
#' @docType data
#'
#' @usage data("pbmcSmall")
#'
#' @format An object of class \code{"dgCMatrix"}.
#'
#' @keywords datasets
#'
#' @source \href{https://support.10xgenomics.com/single-cell-gene-expression/datasets/1.1.0/pbmc3k}{3k PBMCs from a Healthy Donor}
#'
#' @examples
#' data("pbmcSmall")
#' str(pbmcSmall)
#'
#' @importClassesFrom Matrix dgCMatrix
"pbmcSmall"

#' Normalize scRNAseq data
#'
#' @description \code{Dino} removes cell-to-cell variation in observed
#' counts due to the effects of sequencing depth from single-cell mRNA
#' sequencing experiments. \code{Dino} was particularly designed with UMI
#' based protocols in mind, but is applicable to non-UMI based chemistries
#' in the library preparation stage of sequencing.
#'
#' @usage Dino(counts, nCores = 2, prec = 3, minNZ = 10,
#'     nSubGene = 1e4, nSubCell = 1e4, depth = NULL, slope = NULL,
#'     minSlope = 1/2, maxSlope = 2, clusterSlope = TRUE,
#'     returnMeta = FALSE, doRQS = FALSE,
#'     emPar = list(maxIter = 100, tol = 0.1, conPar = 15, maxK = 100), ...)
#'
#' @param counts A numeric matrix object of expression counts - usually in
#'     dgCMatrix format for memory efficiency. Column names denote cells
#'     (samples or droplets) and row names denote genes.
#' @param nCores A non-negative integer scalar denoting the number of cores
#'     which should be used. Setting nCores to 0 uses all cores as determined by
#'     running \code{parallel::detectCores()}
#' @param prec A positive integer denoting the number of decimals to which to
#'     round depth (if estimated internally via \code{depth = NULL}) and
#'     normalized counts for computational efficiency.
#' @param minNZ A positive integer denoting the minimum number of non-zero
#'     counts for a gene to be normalized by the Dino algorithm. It is
#'     recommended to pre-filter the \emph{counts} matrix such that all genes
#'     meet this threshold. Otherwise, genes with fewer than \emph{minNZ}
#'     non-zeros will be scaled by depth for normalization.
#' @param nSubGene A positive integer denoting the number of genes to subset
#'     for calculation of \emph{slope}.
#' @param nSubCell A positive integer denoting the number of samples to subset
#'     for calculation of \emph{slope} and the EM algorithm.
#' @param depth A numeric vector of length equal to the columns of counts.
#'     \emph{depth} denotes a median-centered, log-scale measure of cell-wise
#'     sequencing depth. \code{Dino} defaults to defining depth as the
#'     (within-cell) sum of counts across genes, followed by a log and
#'     median-centering transformation.
#' @param slope A numeric scalar denoting the count-depth relationship on
#'     the log-log scale. Typical values are close to 1 (implying a unit
#'     increase in depth corresponds to a unit increase in expected counts on
#'     the log-log scale), but may be higher, particularly in the case of
#'     non-UMI protocols. \code{Dino} defaults to estimating \emph{slope}
#'     internally.
#' @param minSlope A numeric scalar denoting the minimum slope. Fitted slopes
#'     below this value will return a warning and be set to 1
#' @param maxSlope A numeric scalar denoting the maximum slope. Fitted slopes
#'     above this value will return a warning and be set to 1
#' @param clusterSlope A logical indicating whether cells should be
#'     pre-clustered prior to calculation of slope. Under the default where
#'     cells are pre-clustered, cluster is used as a factor in the regression.
#' @param returnMeta A logical indicating whether metadata (sequencing depth
#'     and slope) should be returned.
#' @param doRQS A logical indicating how normalization resampling is to be done.
#'     By default (F), normalization is done by resampling from the full
#'     posterior distribution. Alternately, restricted quantile sampling (RQS)
#'     can be performed to enforce stronger preservation of expression ranks
#'     in normalized data. Currently RQS is considered experimental.
#' @param emPar A list of parameters to send to the EM algorithm.
#'     \emph{maxIter} denotes the maximum number of model updates. \emph{tol}
#'     denotes the cutoff threshold for reductions in the log likelihood
#'     function. \emph{conPar} denotes the concentration parameter for the
#'     resampling. \emph{conPar = 1} implies full resampling from the fitted
#'     distribution. As \emph{conPar} increases, the normalized expression
#'     converges to the scale-factor normalized values. \emph{maxK} denotes the
#'     maximum number of mixture components in the mixture model.
#' @param ... Additional parameters to pass to \code{Scran::quickCluster}.
#'
#' @return \code{Dino} by default returns a matrix of normalized expression
#'     with identical dimensions as \emph{counts}. If \emph{returnMeta = TRUE},
#'     then \code{Dino} returns a list of normalized expression, sequencing
#'     depth, and slope.
#'
#' @references Brown, J., Ni, Z., Mohanty, C., Bacher, R. and Kendziorski, C.
#'     (2020) "Normalization by distributional resampling of high throughput
#'     single-cell RNA-sequencing data." bioRxiv.
#'     \href{https://doi.org/10.1101/2020.10.28.359901}{https://doi.org/10.1101/2020.10.28.359901}
#'
#' @export
#'
#' @examples
#' # raw data
#' data("pbmcSmall")
#' str(pbmcSmall)
#'
#' # run Dino on raw expression matrix
#' pbmcSmall_Norm <- Dino(pbmcSmall)
#' str(pbmcSmall_Norm)
#'
#' @author Jared Brown
#'
#' @importFrom Matrix Matrix
#' @importFrom Matrix colSums
#' @importFrom Matrix rowSums
#' @importFrom Matrix t
Dino <- function(
    counts, nCores = 2, prec = 3, minNZ = 10, nSubGene = 1e4, nSubCell = 1e4,
    depth = NULL, slope = NULL, minSlope = 1/2, maxSlope = 2,
    clusterSlope = TRUE, returnMeta = FALSE, doRQS = FALSE,
    emPar = list(maxIter = 100, tol = 0.1, conPar = 15, maxK = 100), ...
) {
    ## Perform argument checks
    checkOut <- check_DinoIn(counts, nCores, prec, minNZ,
        nSubGene, nSubCell, depth, slope, returnMeta)
    counts <- t(checkOut$counts)
    nCores <- checkOut$nCores
    prec <- checkOut$prec
    minNZ <- checkOut$minNZ
    nSubGene <- checkOut$nSubGene
    nSubCell <- checkOut$nSubCell
    depth <- checkOut$depth
    slope <- checkOut$slope
    returnMeta <- checkOut$returnMeta


    ## Set depth
    depth <- estDepth_func(depth, counts, prec)
    depthRep <- calcDepthRep(depth)


    ## Calculate slope
    slope <- estSlope_func(
        slope, counts, depth, nSubGene, nSubCell, clusterSlope, nCores, minSlope,
        maxSlope, ...
    )


    ## Resample counts
    normCounts <- dinoResamp_func(
        counts, nSubCell, depth, minNZ, slope, prec, doRQS, emPar, nCores,
        depthRep
    )

    message("Done")
    if(returnMeta == FALSE) {
        return(normCounts)
    } else {
        return(list(
            normMat = normCounts,
            depth = depth,
            slope = slope
        ))
    }
}


# dinoResamp_func is a helper function to resample expression
dinoResamp_func <- function(
    counts, nSubCell, depth, minNZ, slope, prec, doRQS, emPar, nCores,
    depthRep
) {
    if(nrow(counts) > nSubCell) {
        subInd <- sample.int(nrow(counts), nSubCell)
        depthRep <- calcDepthRep(depth[subInd])
    } else {
        subInd <- seq_len(nrow(counts))
    }
    dinoGenes <- which(colSums(counts[subInd, ] > 0) >= minNZ)
    libGenes <- which(!(seq_len(ncol(counts[subInd, ])) %in% dinoGenes))
    if(length(libGenes) > 0){
        warning(cat(
            "Some genes have non-zero expression below",
            "'minNZ' when subsampled\nand will be normalized via scale-factor"
        ))
        libCounts <- counts[, libGenes, drop = FALSE]
        countList <- splitGenes(counts[, dinoGenes, drop = FALSE])
    } else {
        countList <- splitGenes(counts[, dinoGenes, drop = FALSE])
    }
    prll <- setPar(nCores, countList)

    message("Normalizing counts by resampling")
    if(length(libGenes) > 0) {
        normCounts_lib <- libCounts / exp(depth)
        colnames(normCounts_lib) <- colnames(counts)[libGenes]
        normCounts_Dino <- parResampCounts(
            countList, depth, depthRep, slope, prec, subInd, prll, doRQS, emPar
        )
        colnames(normCounts_Dino) <- colnames(counts)[dinoGenes]
        normCounts <- cbind(normCounts_Dino, normCounts_lib)
        normCounts <- normCounts[, colnames(counts)]
    } else {
        normCounts <- parResampCounts(
            countList, depth, depthRep, slope, prec, subInd, prll, doRQS, emPar
        )
    }

    colnames(normCounts) <- colnames(counts)
    rownames(normCounts) <- rownames(counts)
    normCounts <- t(normCounts)

    return(normCounts)
}


# estSlope_func is a helper function to estimate regression slope
estSlope_func <- function(
    slope, counts, depth, nSubGene, nSubCell, clusterSlope, nCores, minSlope,
    maxSlope, ...
) {
    if(is.null(slope)) {
        message("Calculating regression slope")
        colnames(counts) <- unname(colnames(counts))
        slope <- calcSlope(
            counts, depth, nSubGene, nSubCell, clusterSlope, nCores, ...
        )
        if(slope < minSlope) {
            warning(cat(
                "Fitted slope (", round(slope, 3), ") below minSlope (",
                round(minSlope, 3), ") -- setting slope = 1"
            ))
            slope = 1
        }
        if(slope > maxSlope) {
            warning(cat(
                "Fitted slope (", round(slope, 3), ") above maxSlope (",
                round(maxSlope, 3), ") -- setting slope = 1"
            ))
            slope = 1
        }
    }

    return(slope)
}


# estDepth_func is a helper function to estimate sequencing depth
estDepth_func <- function(depth, counts, prec) {
    if(is.null(depth)) {
        message("Computing sequencing depth")
        depth <- log(rowSums(counts))
        depth <- depth - median(depth)
        depth <- round(depth, prec)
    }

    return(depth)
}


#' Create Seurat object from Dino normalized data
#'
#' @description \code{SeuratFromDino} is a wrapper simplifying the export of
#'     \code{Dino} normalized counts to a \emph{Seurat} object for secondary
#'     analysis.
#'
#' @usage SeuratFromDino(counts, doNorm = TRUE, doLog = TRUE, ...)
#'
#' @param counts A numeric matrix of count data, either raw (eg. UMIs) or
#'     normalized expression.
#' @param doNorm A logical indicating whether to normalize the input
#'     \emph{counts} data before exporting results to a \emph{Seurat} object.
#'     By default, it is assumed that the contents of \emph{counts} raw
#'     expression which should be normalized.
#' @param doLog A logical indicating whether normalized counts should be log
#'     transformed with a psuedocount of 1 prior to export.
#' @param ... Further arguments to pass to \emph{Dino}
#'
#' @return \code{SeuratFromDino} returns a Seurat object using Dino normalized
#'     and log transformed expression (default) for downstream analysis in the
#'     Seurat pipeline.
#'
#'     If \emph{returnMeta = T} is passed to \emph{Dino}, then \emph{depth} and
#'     \emph{slope} results are stored in the \emph{Misc} slot under the
#'     names \emph{depth} and \emph{slope} respectively.
#'
#' @references Brown, J., Ni, Z., Mohanty, C., Bacher, R. and Kendziorski, C.
#'     (2020). "Normalization by distributional resampling of high throughput
#'     single-cell RNA-sequencing data." bioRxiv.
#'     \href{https://doi.org/10.1101/2020.10.28.359901}{https://doi.org/10.1101/2020.10.28.359901}
#'
#'     Satija, R., Farrell, J.A., Gennert, D., Schier, A.F. and Regev, A.
#'     (2015). "Spatial reconstruction of single-cell gene expression data."
#'     Nat. Biotechnol., 33, 495–502.
#'     \href{https://doi.org/10.1038/nbt.3192}{https://doi.org/10.1038/nbt.3192}
#'
#' @export
#'
#' @examples
#' # raw data
#' data("pbmcSmall")
#' str(pbmcSmall)
#'
#' # run Dino on raw expression matrix, output Seurat object
#' pbmcSmall_Seurat <- SeuratFromDino(pbmcSmall)
#' str(pbmcSmall_Seurat)
#'
#' @author Jared Brown
#'
#' @importFrom Seurat CreateAssayObject
#' @importFrom Seurat CreateSeuratObject
#' @importFrom Seurat Misc
SeuratFromDino <- function(counts, doNorm = TRUE, doLog = TRUE, ...) {
    if(doNorm) {
        normOut <- Dino(counts, ...)
        argList <- list(...)
        if("returnMeta" %in% names(argList)) {
            if(argList$returnMeta) {
                if(doLog) {
                    normAssay <-
                        CreateAssayObject(data = log1p(normOut$normMat))
                } else {
                    normAssay <-
                        CreateAssayObject(data = normOut$normMat)
                }
                normAssay <- CreateSeuratObject(normAssay)
                Seurat::Misc(normAssay, "depth") <- normOut$depth
                Seurat::Misc(normAssay, "slope") <- normOut$slope
            } else {
                if(doLog) {
                    normAssay <- CreateAssayObject(data = log1p(normOut))
                } else {
                    normAssay <- CreateAssayObject(data = normOut)
                }
                normAssay <- CreateSeuratObject(normAssay)
            }
        } else {
            if(doLog) {
                normAssay <- CreateAssayObject(data = log1p(normOut))
            } else {
                normAssay <- CreateAssayObject(data = normOut)
            }
            normAssay <- CreateSeuratObject(normAssay)
        }
    } else {
        if(doLog) {
            normAssay <- CreateAssayObject(data = log1p(counts))
        } else {
            normAssay <- CreateAssayObject(data = counts)
        }
        normAssay <- CreateSeuratObject(normAssay)
    }
    return(normAssay)
}


#' Run Dino normalization on a SingleCellExperiment dataset
#'
#' @description \code{Dino_SCE} is a wrapper simplifying the application of the
#'     \emph{Dino} method to data formatted as a \emph{SingleCellExperiment}
#'
#' @usage Dino_SCE(SCE, ...)
#'
#' @param SCE A \emph{SingleCellExperiment} object with unnormalized count data
#'     (eg. raw  UMIs) in the \emph{assays} slot under the name \emph{counts}.
#' @param ... Further arguments to pass to \emph{Dino}
#'
#' @return \code{Dino_SCE} returns a \emph{SingleCellExperiment} object using
#'     Dino normalized expression in the \emph{assays} slot under the
#'     \emph{normcounts} name for downstream analysis.
#'
#'     If \emph{returnMeta = T} is passed to \emph{Dino}, then \emph{depth} and
#'     \emph{slope} results are stored in the \emph{metadata} slot under the
#'     names \emph{depth} and \emph{slope} respectively.
#'
#' @references Brown, J., Ni, Z., Mohanty, C., Bacher, R. and Kendziorski, C.
#'     (2020). "Normalization by distributional resampling of high throughput
#'     single-cell RNA-sequencing data." bioRxiv.
#'     \href{https://doi.org/10.1101/2020.10.28.359901}{https://doi.org/10.1101/2020.10.28.359901}
#'
#'     Amezquita, R.A., Lun, A.T.L., Becht, E., Carey, V.J., Carpp, L.N.,
#'     Geistlinger, L., Marini, F., Rue-Albrecht, K., Risso, D., Soneson, C.,
#'     et al. (2020). "Orchestrating single-cell analysis with Bioconductor."
#'     Nat. Methods, 17, 137–145.
#'     \href{https://doi.org/10.1038/s41592-019-0654-x}{https://doi.org/10.1038/s41592-019-0654-x}
#'
#' @export
#'
#' @examples
#' # raw data
#' data("pbmcSmall")
#' str(pbmcSmall)
#'
#' # format as SingleCellExperiment
#' library(SingleCellExperiment)
#' pbmc_SCE <- SingleCellExperiment(assays = list("counts" = pbmcSmall))
#'
#' # Run Dino
#' pbmc_SCE <- Dino_SCE(pbmc_SCE)
#' str(pbmc_SCE)
#' str(normcounts(pbmc_SCE))
#'
#' @author Jared Brown
#'
#' @import SingleCellExperiment
#' @importFrom SummarizedExperiment assayNames
#' @importFrom methods is
#' @importFrom S4Vectors metadata
Dino_SCE <- function(SCE, ...) {
    if(!is(SCE, "SingleCellExperiment")) {
        stop("'SCE' is not a SingleCellExperiment")
    }
    if(is.null(assayNames(SCE))) {
        stop("no 'counts' assay present")
    }
    if(!("counts" %in% assayNames(SCE))) {
        stop("no 'counts' assay present")
    }

    argList <- list(...)
    if("returnMeta" %in% names(argList)) {
        if(argList$returnMeta) {
            countMat <- counts(SCE)
            normOut <- Dino(countMat, ...)
            SingleCellExperiment::normcounts(SCE) <- normOut$normMat
            S4Vectors::metadata(SCE)[["depth"]] <- normOut$depth
            S4Vectors::metadata(SCE)[["slope"]] <- normOut$slope
        } else {
            countMat <- counts(SCE)
            SingleCellExperiment::normcounts(SCE) <- Dino(countMat, ...)
        }
    } else {
        countMat <- counts(SCE)
        SingleCellExperiment::normcounts(SCE) <- Dino(countMat, ...)
    }

    return(SCE)
}

data("pbmcSmall")
library(SingleCellExperiment)

test_that("Test Dino_SCE", {
    pbmc_SCE <- SingleCellExperiment(assays = list("counts" = pbmcSmall))

    expect_that(Dino_SCE(pbmc_SCE, nCores = 2), is_a("SingleCellExperiment"))
})

data("pbmcSmall")

test_that("Test Dino 'counts' input", {
    # Commented out to save time for Bioconductor submission

    expect_error(Dino(seq_len(10)),
                 "'counts' not a matrix; dim returns null")

    expect_error(Dino(pbmcSmall - 1),
                 "'counts' contains negative values")

    expect_error(Dino(cbind(pbmcSmall, 0)),
                 "'counts' contains samples with all-zero genes")
})

test_that("Test Dino 'minNZ' input", {
    expect_error(Dino(pbmcSmall, minNZ = seq_len(10)),
                 "'minNZ' doesn't have length 1")

    expect_error(Dino(pbmcSmall, minNZ = 1),
                 "'minNZ' less than 2, choose a larger value")

    expect_error(Dino(pbmcSmall[, seq_len(5)]),
                 "all genes have fewer than 'minNZ' non-zero values")
})

test_that("Test Dino 'depth' input", {
    expect_error(Dino(pbmcSmall[, seq_len(20)], depth = log(colSums(pbmcSmall))),
                 "'depth' has length not equal to the number of samples")
})

# calcSlope is a helper function to calculate the expression-depth slope
#
#' @importFrom Matrix Matrix
#' @importFrom Matrix colSums
#' @importFrom scran quickCluster
#' @importFrom BiocSingular IrlbaParam
#' @importFrom stats density
#' @importFrom stats approxfun
calcSlope <- function(counts, depth, nSubGene, nSubCell, clusterSlope,
                nCores, ...) {
    if(nrow(counts) > nSubCell) {
        subInd <- sample.int(nrow(counts), nSubCell)
        subCounts <- counts[subInd, ]
        subDepth <- depth[subInd]
    } else {
        subCounts <- counts
        subDepth <- depth
    }
    prll <- setPar(nCores)
    if(clusterSlope) {
        clustVec <- quickCluster(x = t(subCounts), BSPARAM = IrlbaParam(),
                                    BPPARAM = prll, ...)
    } else {
        clustVec <- rep(1, nSubCell)
    }

    # Subset genes with a minimum proportion of non-zeros
    subGenes <- colSums(subCounts > 0) / nrow(subCounts) >= 1e-2
    geneMu <- log(apply(subCounts, 2, function(x) {
        exp(mean(log(x + 1))) - 1
    }))
    muDens <- density(geneMu)
    muFunc <- approxfun(muDens$x, muDens$y)
    geneSamp <- sample(which(subGenes),
                    min(sum(subGenes), nSubGene),
                    prob = 1 / muFunc(geneMu[subGenes]))

    # Fit to batches of genes
    datList <- splitGenes(subCounts[, geneSamp], nCol = 40)
    clustTab <- table(clustVec)
    prll <- setPar(nCores, datList)
    slopeVec <- parPoisSlope(datList, subDepth, clustVec, clustTab,
                    prll, pctNZ = 0.01)

    return(mean(slopeVec))
}


# poisSlope is a helper function to compute poisson regression slopes
#
#' @importFrom stats glm
poisSlope <- function(y, depth, clustVec, clustTab, pctNZ = 0.01) {
    clustNZ <- table(clustVec[y > 0])
    regInd <- clustVec %in% names(clustNZ)[clustNZ > 1]
    y <- y[regInd]; depth <- depth[regInd]; clustVec <- clustVec[regInd]
    clustVec <- factor(clustVec, levels = unique(clustVec))
    if(length(levels(clustVec)) >= 2) {
        return(glm(y ~ depth + clustVec, family = "poisson")$coefficients[2])
    } else {
        return(glm(y ~ depth, family = "poisson")$coefficients[2])
    }
}


# parPoisSlope is a helper function to compute poisson regression slopes
#
#' @importFrom Matrix Matrix
parPoisSlope <- function(datList, depth, clustVec, clustTab, prll,
                    pctNZ = 0.01) {
    slopeVec <- bplapply(datList,
                    function(subDat, depth, clustVec, clustTab, pctNZ) {
                        subSlopes <- rep(1, ncol(subDat))
                        for(i in seq_len(ncol(subDat))) {
                            subSlopes[i] <- poisSlope(subDat[, i], depth,
                                                clustVec, clustTab,
                                                pctNZ)
                        }
                        return(subSlopes)
                    }, depth = depth, clustVec = clustVec,
                    clustTab = clustTab, pctNZ = pctNZ, BPPARAM = prll)
    do.call(c, slopeVec)
}

# calcDev is a helper function to calculate upper and lower sums of absolute
# deviations for quantile regression.
#
#' @importFrom grDevices chull
#' @importFrom stats median
#' @importFrom stats approxfun
#' @importFrom stats optimize
#' @importFrom stats pgamma
#' @importFrom stats qgamma
calcDev <- function(y, depth, depthRep, slope = 1) {
    # Initialize return
    ret <- devInit_func(y, depth, depthRep, slope)

    # Convexify percentile estimates #
    ret <- convexDev_func(ret, slope)

    # Subset absolute deviations to reduce memory
    if(nrow(ret) > 100) {
        subLen <- floor(log(nrow(ret)))
        subInd <- seq_len(ceiling(nrow(ret) / subLen)) * subLen - (subLen - 1)
        ret <- ret[subInd[seq_len(length(subInd) - subLen)], ]
    }
    ret <- ret[, 6:7]

    #  Estimate eCDF from 0 to minimum estimated percentile
    ret <- estLowerCDF_func(ret)

    return(ret)
}


# devInit_func is a helper function to initialize eCDF estimation
devInit_func <- function(y, depth, depthRep, slope) {
    # Calculate log y (pseudo-count at 0s which preserves rank order)
    logY <- pmax(log(y), log(0.999))

    ret <- data.frame(
        intercept = sort(depth - (logY - log(0.999)) / slope)
    )
    intDup <- duplicated(ret$intercept)
    freq <- c(which(!intDup), length(depth) + 1)
    ret <- ret[!intDup, , drop = FALSE]
    ret$freq <- freq[-1] - freq[-length(freq)]

    # Calculate lower deviations
    ret$lowDev <- sum(
        (log(0.999) + slope * (depth - ret$intercept[1])) - logY
    ) - cumsum(
        c(0, diff(ret$intercept) * slope * rev(cumsum(rev(ret$freq[-1]))))
    )

    # Calculate upper deviations
    intDepthOrd <- order(c(ret$intercept, depthRep$depth))
    ret$highDev <- c(
        0,
        cumsum(
            (
                cumsum(c(ret$freq, -depthRep$rep)[intDepthOrd])
                [-(nrow(ret) + nrow(depthRep))] *
                    diff(c(ret$intercept, depthRep$depth)[intDepthOrd])
            ) * slope
        )[intDepthOrd <= nrow(ret)][-nrow(ret)]
    )

    # Calculate numbers of zeros (censored points)
    zInd <- match(depth[y == 0], ret$intercept)
    ret$zInt <- 0
    ret$zInt[sort(unique(zInd))] <- table(zInd)

    # Remove duplicate quantiles
    if(ret$freq[1] == ret$zInt[1]) {
        ret <- ret[which.max(ret$freq > ret$zInt):nrow(ret), ]
    }

    return(ret)
}


# convexDev_func is a helper function to convexify the estimated absolute
# devaitions
convexDev_func <- function(ret, slope) {
    nY <- sum(ret$freq)
    chullInt <- chull(ret$highDev, ret$lowDev)
    chullInt <- chullInt[
        which.max(chullInt == nrow(ret)):which.max(chullInt == 1)
    ]
    hRet <- ret[rev(chullInt), ]
    hRet$m <- c(
        1e3 * diff(hRet$lowDev[seq_len(2)]) /
            diff(hRet$highDev[seq_len(2)]),
        diff(hRet$lowDev) / diff(hRet$highDev)
    )
    mFunc <- approxfun(x = hRet$highDev, y = hRet$m)
    m <- mFunc(ret$highDev)
    ret$pct <- -m / (1 - m)
    pctDiff <- pmax(-diff(ret$pct), (ret$freq - ret$zInt)[-nrow(ret)] / nY)
    lZ <- ret$zInt[1]
    for(i in 2:length(pctDiff)) {
        lZ <- lZ + ret$zInt[i] -
            (pctDiff[i-1] - (ret$freq[i - 1] - ret$zInt[i - 1]) / nY) * nY
        pctDiff[i] <- min(pctDiff[i], (ret$freq[i] - ret$zInt[i] + lZ) / nY)
    }
    ret$pct <- c(1, 1 - cumsum(pctDiff))
    ret$quant <- -ret$intercept * slope + log(0.999)
    ret <- ret[!duplicated(ret$pct), ]
    ret <- ret[!duplicated(ret$quant), ]

    return(ret)
}


# estLowerCDF_func is a helper function to estimate the lower bound of the  eCDF
estLowerCDF_func <- function(ret) {
    quantFun <- approxfun(ret$pct, ret$quant)
    # Approximate mean from eCDF (1,000 bins)
    mu <- sum(
        exp(quantFun(seq(min(ret$pct), 1, length = 1e3)))  *
            ((1 - min(ret$pct)) / 1e3)
    )
    minQ <- exp(min(ret$quant))
    minP <- min(ret$pct)
    # Estimate lower bound gamma distribution with sannity controls
    out <- optimize(f = function(beta, mu, minQ, minP) {
        (pgamma(minQ, mu / beta, beta) - minP)^2
    }, interval = c(0, 1e2), mu = mu, minQ = minQ, minP = minP, tol = 1e-6)
    beta <- out$minimum

    minP <- min(minP, pgamma(minQ, mu / beta, beta))

    # Append eCDF lower bound with 100 fitted gamma points
    ret <- rbind(
        ret,
        data.frame(
            pct = seq(minP, 0, length = ceiling(100 * minP)),
            quant = log(qgamma(
                seq(minP, 0, length = ceiling(100 * minP)), mu / beta, beta
            )) - (min(ret$quant) - log(qgamma(minP, mu / beta, beta)))
        )
    )
    ret <- ret[!duplicated(ret$pct), ]
    ret <- ret[!is.infinite(ret$quant), ]
    ret <- ret[!duplicated(ret$quant), ]
    if(min(ret$pct) != 0) {
        ret <- rbind(
            ret, c(0, min(ret$quant) - 100)
        )
    }

    return(ret)
}

# check_DinoIn is a helper function to verify the correct format of the
# arguments of Dino
#
#' @importFrom Matrix Matrix
check_DinoIn <- function(counts, nCores, prec, minNZ,
                    nSubGene, nSubCell, depth, slope, returnMeta) {
    ## Checks
    counts <- checkCounts(counts)
    minNZ <- checkMinNZ(minNZ, counts)
    checkRet <- checkPrecDepth(prec, depth, counts)
    prec <- checkRet$prec
    depth <- checkRet$depth
    nSubGene <- checkNSubGene(nSubGene)
    nSubCell <- checkNSubCell(nSubCell)
    nCores <- checkNCores(nCores)
    sope <- checkSlope(slope)
    returnMeta <- checkRetM(returnMeta)
    return(list(counts = counts, nCores = nCores, prec = prec,
                minNZ = minNZ, nSubGene = nSubGene, nSubCell = nSubCell,
                depth = depth, slope = slope, returnMeta = returnMeta))
}


# checkCounts is a helper function to verify the correct format of the counts
# argument
#
#' @importFrom Matrix Matrix
#' @importFrom Matrix colSums
#' @importFrom methods is
checkCounts <- function(counts) {
    if(is.null(dim(counts))) {
        stop("'counts' not a matrix; dim returns null")
    }
    if(!is(counts, "dgCMatrix")) {
        message("converting 'counts' to dgCMatrix")
        counts <- as(counts, "dgCMatrix")
    }
    if(any(round(counts) != counts)) {
        message("rounding 'counts'")
        counts <- Matrix(round(counts), sparse = TRUE)
    }
    if(any(counts < 0)) {
        stop("'counts' contains negative values")
    }
    if(any(colSums(counts) == 0)) {
        stop("'counts' contains samples with all-zero genes")
    }
    return(counts)
}


# checkMinNZ is a helper function to verify the correct format of the minNZ
# argument
#
#' @importFrom Matrix Matrix
#' @importFrom Matrix rowSums
checkMinNZ <- function(minNZ, counts) {
    if(!is.numeric(minNZ)) {
        stop("'minNZ' is not a number")
    }
    if(!is.null(dim(minNZ))) {
        stop("'minNZ' has dimension greater than 1")
    }
    if(length(minNZ) > 1) {
        stop("'minNZ' doesn't have length 1")
    }
    if(minNZ != round(minNZ)) {
        message("rounding 'minNZ'")
        minNZ <- round(minNZ)
    }
    if(minNZ < 2) {
        stop("'minNZ' less than 2, choose a larger value")
    }
    if(minNZ < 10) {
        warning("'minNZ' less than 10, this is not recommended")
    }
    if(!any(rowSums(counts > 0) >= minNZ)) {
        stop("all genes have fewer than 'minNZ' non-zero values")
    }
    return(minNZ)
}


# checkPrecDepth is a helper function to verify the correct format of the prec
# and depth arguments
#
#' @importFrom Matrix Matrix
checkPrecDepth <- function(prec, depth, counts) {
    if(!is.numeric(prec)) {
        stop("'prec' is not a number")
    }
    if(!is.null(dim(prec))) {
        stop("'prec' has dimension greater than 1")
    }
    if(length(prec) > 1) {
        stop("'prec' doesn't have length 1")
    }
    if(prec != round(prec)) {
        message("rounding 'prec'")
        prec <- round(prec)
    }

    if(!is.null(depth))  {
        if(!is.numeric(depth)) {
            stop("'depth' is not of type numeric")
        }
        if(!is.null(dim(depth))) {
            stop("'depth' has dimension greater than 1")
        }
        if(length(depth) != ncol(counts)) {
            stop("'depth' has length not equal to the number of samples")
        }
    }

    return(list(
        prec = prec,
        depth = depth
    ))
}


# checkNSubGene is a helper function to verify the nSubGene argument
checkNSubGene <- function(nSubGene) {
    if(!is.numeric(nSubGene)) {
        stop("'nSubGene' is not of type numeric")
    }
    if(!is.null(dim(nSubGene))) {
        stop("'nSubGene' has dimension greater than 1")
    }
    if(nSubGene < 100) {
        warning("'nSubGene' less than 100, raise threshold before re-running")
    }
    if(nSubGene != round(nSubGene)) {
        message("rounding 'nSubGene'")
        nSubGene <- round(nSubGene)
    }
    return(nSubGene)
}


# checkNSubCell is a helper function to verify the correct format of the
# nSubCell argument
checkNSubCell <- function(nSubCell) {
    if(!is.numeric(nSubCell)) {
        stop("'nSubCell' is not of type numeric")
    }
    if(!is.null(dim(nSubCell))) {
        stop("'nSubCell' has dimension greater than 1")
    }
    if(nSubCell < 100) {
        warning("'nSubCell' less than 100, raise threshold before re-running")
    }
    if(nSubCell != round(nSubCell)) {
        message("rounding 'nSubCell'")
        nSubCell <- round(nSubCell)
    }
    return(nSubCell)
}


# checkNCores is a helper function to verify the correct format of the nCores
# argument
#
#' @importFrom parallel detectCores
checkNCores <- function(nCores) {
    if(!is.numeric(nCores)) {
        stop("'nCores' is not a number")
    }
    if(!is.null(dim(nCores))) {
        stop("'nCores' has dimension greater than 1")
    }
    if(length(nCores) > 1) {
        stop("'nCores' doesn't have length 1")
    }
    if(nCores != round(nCores)) {
        message("rounding 'nCores'")
        nCores <- round(nCores)
    }
    if(nCores < 0) {
        stop("'nCores' is less than 0")
    }
    if(nCores == 0) {
        message("using all available cores")
        nCores <- detectCores()
    }
    if(nCores > detectCores()) {
        warning(cat("'nCores' is greater than the number of available ",
                    "cores, reducing to ", detectCores()))
        nCores <- detectCores()
    }
    return(nCores)
}


# checkSlope is a helper function to verify the correct format of the slope
# argument
checkSlope <- function(slope) {
    if(is.null(slope)) {
        return(NULL)
    }
    if(!is.numeric(slope)) {
        stop("'slope' is not of type numeric")
    }
    if(!is.null(dim(slope))) {
        stop("'slope' has dimension greater than 1")
    }
    if(slope < 0.25 | slope > 4) {
        warning("'slope' is outside of usual bounds for sequencing experiments")
    }
    return(slope)
}


# returnMeta is a helper function to verify the correct format of the
# returnMeta argument
checkRetM <- function(returnMeta) {
    if(!is.logical(returnMeta)) {
        stop("'returnMeta' is not of type logical")
    }
    if(length(returnMeta) > 1) {
        stop("'returnMeta' has length greater than 1")
    }
    return(returnMeta)
}

# resampCounts is a helper function to normalize counts by resampling from an
# estimated expression distribution conditional on the observed counts
#
#' @importFrom Matrix Matrix
#' @importFrom stats approxfun
#' @importFrom stats bw.ucv
#' @importFrom stats bw.nrd
#' @importFrom stats ppois
#' @importFrom stats qgamma
#' @importFrom stats pgamma
#' @importFrom stats runif
#' @importFrom stats rgamma
resampCounts <- function(countMat, depth, depthRep, slope,
                    prec, subInd, doRQS, emPar) {
    subNorm <- apply(countMat, 2,
                function(y, depth, depthRep, slope, prec, subInd) {
        # Initialize EM parameters
        parList <- emParInit_func(y, depth, depthRep, slope, subInd, emPar)
        fitSlope <- parList$fitSlope
        parList <- parList$parList

        # Fit EM parameters
        parList <- fitEM_func(parList, emPar, subInd)

        # Resample normalized expression values
        if(doRQS) {
            return(rqsSample_func(parList, y, depth, fitSlope, prec))
        } else {
            return(posteriorSample_func(
                parList, y, depth, fitSlope, prec, emPar
            ))
        }
    }, depth = depth, depthRep = depthRep, slope = slope,
    prec = prec, subInd = subInd)
    subNorm <- Matrix(subNorm, sparse = TRUE)
    return(subNorm)
}


# parResampCounts is a helper function to parallelize the  resampling algorithm
#
#' @importFrom BiocParallel bplapply
#' @importFrom Matrix Matrix
parResampCounts <- function(countList, depth, depthRep, slope,
                        prec, subInd, prll, doRQS, emPar) {
    normList <- bplapply(countList,
                    function(subCounts, depth, depthRep, slope, prec, subInd) {
                        resampCounts(subCounts, depth, depthRep,
                            slope, prec, subInd, doRQS, emPar)
                    }, depth = depth, depthRep = depthRep,
                    slope = slope, prec = prec, subInd = subInd,
                    BPPARAM = prll)
    while(length(normList) > 2) {
        for(i in seq_len(floor(length(normList) / 2))) {
            if(length(normList) >= 2 * i) {
                normList[[i]] <- cbind(normList[[2 * i - 1]], normList[[2 * i]])
            } else {
                normList[[i]] <- normList[[2 * i - 1]]
            }
        }
        if(length(normList) > 2 * i) {
            i <- i + 1
            normList[[i]] <- normList[[2 * i - 1]]
        }
        while(length(normList) > i) {
            normList[[i + 1]] <- NULL
        }
    }
    if(length(normList) == 2) {
        normList <- cbind(normList[[1]], normList[[2]])
    } else {
        normList <- normList[[1]]
    }
    return(normList)
}


# lLik_func is a helper function to compute the EM log-likelihood up to an
# additive constant
#
#' @importFrom matrixStats rowMaxs
lLik_func <- function(parList) {
    M <- outer(em_y(parList), em_betaVec(parList))
    M <- M - outer(em_depth(parList), exp(em_betaVec(parList)))
    M <- M +
        outer(rep(1, length(em_depth(parList))), em_gamList(parList)$gamVec)
    S <- rowMaxs(M)
    return(sum(log(rowSums(exp(M - S))) + S) -
        length(em_y(parList)) * log(em_gamList(parList)$gamDot))
}


# lLikFull_func is a helper function to compute the EM log-likelihood up to an
# additive constant
#
#' @importFrom matrixStats rowMaxs
lLikFull_func <- function(parList) {
    M <- outer(em_y(parList), em_betaVec(parList))
    M <- M - outer(em_depth(parList), exp(em_betaVec(parList)))
    M <- M +
        outer(rep(1, length(em_depth(parList))), em_gamList(parList)$gamVec)
    S <- rowMaxs(M)
    return(sum(log(rowSums(exp(M - S))) + S) -
            length(em_y(parList)) * log(em_gamList(parList)$gamDot) +
            sum(
                em_y(parList) * log(em_depth(parList)) -
                    lfactorial(em_y(parList))
            ))
}


# lLikUpdate_func is a helper function to compute the EM log-likelihood up to
# an addative constant at update pars
#
#' @importFrom matrixStats rowMaxs
lLikUpdate_func <- function(parList) {
    M <- outer(em_y(parList), em_parUpdate(parList)$betaVec)
    M <- M - outer(em_depth(parList), exp(em_parUpdate(parList)$betaVec))
    M <- M + outer(rep(1, length(em_depth(parList))),
                    em_parUpdate(parList)$gamList$gamVec)
    S <- rowMaxs(M)
    return(sum(log(rowSums(exp(M - S))) + S) -
            length(em_y(parList)) * log(em_parUpdate(parList)$gamList$gamDot))
}


# lpMat_func is a helper function to compute the EM log membership probabilities
#
#' @importFrom matrixStats rowMaxs
#' @importFrom matrixStats colMaxs
lpMat_func <- function(parList) {
    lpMat <- outer(em_betaVec(parList), em_y(parList))
    lpMat <- lpMat - outer(exp(em_betaVec(parList)), em_depth(parList))
    lpMat <- lpMat + em_gamList(parList)$gamVec
    lpMat <- lpMat - outer(rep(1, em_J(parList)), colMaxs(lpMat))
    lpMat <- lpMat - outer(rep(1, em_J(parList)), log(colSums(exp(lpMat))))
    lpMax <- rowMaxs(lpMat)
    return(list(
        spMat = exp(lpMat - lpMax),
        lpMax = lpMax
    ))
}


# lpMatUpdate_func is a helper function to compute the EM log membership
# probabilities at update pars
#
#' @importFrom matrixStats rowMaxs
#' @importFrom matrixStats colMaxs
lpMatUpdate_func <- function(parList) {
    lpMat <- outer(em_parUpdate(parList)$betaVec, em_y(parList))
    lpMat <- lpMat -
        outer(exp(em_parUpdate(parList)$betaVec), em_depth(parList))
    lpMat <- lpMat + em_parUpdate(parList)$gamList$gamVec
    lpMat <- lpMat - outer(rep(1, em_J(parList)), colMaxs(lpMat))
    lpMat <- lpMat - outer(rep(1, em_J(parList)), log(colSums(exp(lpMat))))
    lpMax <- rowMaxs(lpMat)
    return(list(
        spMat = exp(lpMat - lpMax),
        lpMax = lpMax
    ))
}


# gamUpdate_func is a helper function to compute the EM update to the gamma
# parameter
gamUpdate_func <- function(parList) {
    gamUpdate <- log(rowSums(em_pList(parList)$spMat)) +
        em_pList(parList)$lpMax
    return(gamUpdate - gamUpdate[em_zInd(parList)])
}


# gamUpdate2_func is a helper function to compute the EM update to the gamma
# parameter at update pars
gamUpdate2_func <- function(parList) {
    gamUpdate <- log(rowSums(em_pListUpdate(parList)$spMat)) +
        em_pListUpdate(parList)$lpMax
    return(gamUpdate - gamUpdate[em_zInd(parList)])
}


# betaUpdate_func is a helper function to compute the EM update to the beta
# parameter
betaUpdate_func <- function(parList) {
    betaUpdate <- log(pmax(em_pList(parList)$spMat %*% em_y(parList),
                            sqrt(.Machine$double.xmin))) -
        log(pmax(em_pList(parList)$spMat %*% em_depth(parList),
                    sqrt(.Machine$double.xmin)))
    return(c(betaUpdate))
}


# betaUpdate2_func is a helper function to compute the EM update to the beta
# parameter at update pars
betaUpdate2_func <- function(parList) {
    betaUpdate <- log(pmax(em_pListUpdate(parList)$spMat %*% em_y(parList),
                            sqrt(.Machine$double.xmin))) -
        log(pmax(em_pListUpdate(parList)$spMat %*% em_depth(parList),
                    sqrt(.Machine$double.xmin)))
    return(c(betaUpdate))
}


# gTilde_func is a helper function to compute the EM parameter update deltas
gTilde_func <- function(parList) {
    gamUpdate <- gamUpdate_func(parList)
    betaUpdate <- betaUpdate_func(parList)
    return(list(
        gamTilde = gamUpdate - em_gamList(parList)$gamVec,
        betaTilde = betaUpdate - em_betaVec(parList),
        gTilde = c(gamUpdate - em_gamList(parList)$gamVec,
                    betaUpdate - em_betaVec(parList))
    ))
}


# gTildeUpdate_func is a helper function to compute the EM parameter update
# deltas at update pars
gTildeUpdate_func <- function(parList) {
    gamUpdate <- gamUpdate2_func(parList)
    betaUpdate <- betaUpdate2_func(parList)
    return(list(
        gamTilde = gamUpdate - em_parUpdate(parList)$gamList$gamVec,
        betaTilde = betaUpdate - em_parUpdate(parList)$betaVec,
        gTilde = c(gamUpdate - em_parUpdate(parList)$gamList$gamVec,
                    betaUpdate - em_parUpdate(parList)$betaVec)
    ))
}


# gGamma_func is a helper function to compute the likelihood gradient in gamma
gGamma_func <- function(parList) {
    M <- rowSums(em_pList(parList)$spMat) * exp(em_pList(parList)$lpMax)
    M <- M - sum(M) *
        exp(em_gamList(parList)$gamVec - log(em_gamList(parList)$gamDot))
    M[em_zInd(parList)] <- 0
    return(M)
}


# gGamma2_func is a helper function to compute the likelihood gradient in
# gamma at update pars
gGamma2_func <- function(parList) {
    M <- rowSums(em_pListUpdate(parList)$spMat) *
        exp(em_pListUpdate(parList)$lpMax)
    M <- M - sum(M) * exp(em_parUpdate(parList)$gamList$gamVec -
                            log(em_parUpdate(parList)$gamList$gamDot))
    M[em_zInd(parList)] <- 0
    return(M)
}


# gBeta_func is a helper function to compute the likelihood gradient in beta
gBeta_func <- function(parList) {
    M <- em_pList(parList)$spMat %*% em_y(parList) -
        rowSums(em_pList(parList)$spMat *
                outer(exp(em_betaVec(parList)), em_depth(parList)))
    return(c(exp(em_pList(parList)$lpMax) * M))
}


# gBeta2_func is a helper function to compute the likelihood gradient in beta
# at update pars
gBeta2_func <- function(parList) {
    M <- em_pListUpdate(parList)$spMat %*% em_y(parList) -
        rowSums(em_pListUpdate(parList)$spMat *
                outer(exp(em_parUpdate(parList)$betaVec), em_depth(parList)))
    return(c(exp(em_pListUpdate(parList)$lpMax) * M))
}


# parGrad_func is a helper function to compute the EM parameter gradients
parGrad_func <- function(parList) {
    gGam <- gGamma_func(parList)
    gBeta <- gBeta_func(parList)
    return(list(
        gGam = gGam,
        gBeta = gBeta,
        g = c(gGam, gBeta)
    ))
}


# parGradUpdate_func is a helper function to compute the EM parameter
# gradients at update pars
parGradUpdate_func <- function(parList) {
    gGam <- gGamma2_func(parList)
    gBeta <- gBeta2_func(parList)
    return(list(
        gGam = gGam,
        gBeta = gBeta,
        g = c(gGam, gBeta)
    ))
}


# alpha_func is a helper function to compute the initial EM step length
alpha_func <- function(parList) {
    return(c(sign(t(em_d(parList)) %*% em_g(parList)$g)))
}


# parUpdate_func is a helper function to compute the EM parameter updates
parUpdate_func <- function(parList) {
    parVec <- c(em_gamList(parList)$gamVec, em_betaVec(parList)) +
        em_alpha(parList) * em_d(parList)
    return(list(
        gamList = list(
            gamVec = parVec[seq_len(em_J(parList))],
            gamDot = sum(exp(parVec[seq_len(em_J(parList))]))
            ),
        betaVec = parVec[(em_J(parList) + 1):(2 * em_J(parList))],
        parVec = parVec
    ))
}


# emParInit_func is a helper function to initialize the EM parameters from
# raw cout data
emParInit_func <- function(y, depth, depthRep, slope, subInd, emPar) {
    # Calculate eCDF
    outDev <- calcDev(y[subInd], depth[subInd], depthRep, slope)
    cdfList <- list(
        eCDF_fun = approxfun(
            x = outDev$quant,
            y = outDev$pct,
            yleft = 0, yright = 1
        ),
        eQuant_fun = approxfun(
            x = outDev$pct,
            y = outDev$quant,
            yleft = -Inf, yright = max(outDev$quant)
        )
    )

    # Bound slope to exp(-3) times minimum depth
    minD <- min(depth)
    if(minD < 0) {
        fitSlope <- min(slope, (exp(minD - 3) - 1) / (exp(minD) - 1))
    } else {
        fitSlope <- max(slope, (exp(minD - 3) - 1) / (exp(minD) - 1))
    }

    # Initialize EM parameters with 5 strict EM steps
    parList <- initEM_func(
        y = y[subInd],
        slope = fitSlope,
        subCDF = cdfList,
        depth = depth[subInd],
        stepInit = 5,
        emPar
    )

    return(list(
        parList = parList,
        fitSlope = fitSlope
    ))
}


# initEM_func is a helper function to compute the EM parameter updates
#' @importFrom methods new
initEM_func <- function(y, slope, subCDF, depth, stepInit, emPar) {
    # Bound slope to exp(-3) times minimum depth
    minD <- min(depth)
    if(minD < 0) {
        slope <- min(slope, (exp(minD - 3) - 1) / (exp(minD) - 1))
    } else {
        slope <- max(slope, (exp(minD - 3) - 1) / (exp(minD) - 1))
    }
    parList <- new(
        "emFitPars", y = y, depth = exp(depth) * slope - slope + 1,
        J = min(emPar$maxK, max(2, ceiling(sqrt(sum(y > 0)))))
    )
    parList <- em_gamListSet(parList, list(
        gamVec = rep(0, em_J(parList)),
        gamDot = sum(exp(rep(0, em_J(parList))))
    ))
    parList <- em_zIndSet(parList, which.max(em_gamList(parList)$gamVec))
    # bound initial beta values within eCDF minus 1e-4 on either side
    lBound <- max(1e-4, subCDF$eCDF_fun(-100))
    parList <- em_betaVecSet(parList, subCDF$eQuant_fun(
        seq(lBound, 1 - 1e-4, length = em_J(parList))
    ))

    for(i in seq_len(stepInit)) {
        parList <- em_pListSet(parList, lpMat_func(parList))
        parList <- em_gTildeSet(parList, gTilde_func(parList))
        gamList <- em_gamList(parList)
        gamList$gamVec <- em_gamList(parList)$gamVec +
            em_gTilde(parList)$gamTilde
        parrList <- em_gamListSet(parList, gamList)
        parList <- em_zIndSet(parList, which.max(em_gamList(parList)$gamVec))
        gamList <- em_gamList(parList)
        gamList$gamDot <- sum(exp(em_gamList(parList)$gamVec))
        parList  <- em_gamListSet(parList, gamList)
        parList <- em_betaVecSet(
            parList, em_betaVec(parList) + em_gTilde(parList)$betaTilde
        )
    }

    parList <- em_zIndSet(parList, which.max(em_gamList(parList)$gamVec))

    parList <- em_pListSet(parList, lpMat_func(parList))
    parList <- em_gSet(parList, parGrad_func(parList))
    parList <- em_gTildeSet(parList, gTilde_func(parList))
    parList <- em_SSet(parList, matrix(0, 2 * em_J(parList), 2 * em_J(parList)))

    parList <- em_lLikSet(parList, lLik_func(parList))

    return(parList)
}


# fitEM_func is a helper function to fit the EM parameter estimates
fitEM_func <- function(parList, emPar, subInd) {
    nUpdate <- emPar$maxIter
    deltaStop <- emPar$tol
    iter <- 0

    while(iter < nUpdate) {
        iter <- iter + 1

        parList <- em_dSet(
            parList, as.numeric(
                -em_gTilde(parList)$gTilde + em_S(parList) %*% em_g(parList)$g
            )
        )
        parList <- em_alphaSet(parList, -1)

        # Sanity check on bounds of k+1 parameter estimates
        parList <- emParCheck_func(parList, subInd)
        if(parList$endLoop) {
            parList <- parList$parList
            break
        } else {
            parList <- parList$parList
        }

        # Accelerated EM step
        parList <- takeEMStep_func(parList)

        # Check end condition
        if(em_lLikUpdate(parList) - em_lLik(parList) <= deltaStop) {
            break
        }
        parList <- em_lLikSet(parList, em_lLikUpdate(parList))
    }

    return(parList)
}


# emParCheck_func is a helper function to perform sanity checks on  projected
# EM update values
emParCheck_func <- function(parList, subInd) {
    # Sanity check on bounds of k+1 parameter estimates
    endLoop <- FALSE

    gamInd <- em_gamList(parList)$gamVec +
        em_alpha(parList) * em_d(parList)[seq_len(em_J(parList))]
    if(!all(gamInd >= -75)) {
        if(em_J(parList) > 2) {
            parList <- parListUpdate_func(parList)
            parList <- em_lLikSet(parList, lLik_func(parList))
            parList <- em_SSet(parList, em_S(parList) * 0)
        } else {endLoop <- TRUE}
    }

    betaInd <- em_betaVec(parList) +
        em_alpha(parList) *
        em_d(parList)[(em_J(parList) + 1):(2 * em_J(parList))]
    if(!all(betaInd > -75)) {
        if(em_J(parList) > 2) {
            parList <- parListUpdate_func(parList, doBeta = TRUE)
            parList <- em_lLikSet(parList,  lLik_func(parList))
            parList <- em_SSet(parList, em_S(parList) * 0)
        } else {endLoop <- TRUE}
    }
    if(
        any(gamInd > 75) | any(betaInd > 75) |
        max(gamInd) - min(gamInd) > log(2 * length(subInd))
    ) {
        parList <- em_SSet(parList, em_S(parList) * 0)
    }

    return(list(
        parList = parList,
        endLoop = endLoop
    ))
}


# parListUpdate_func is a helper function to update parList when a parameter
# is to be removed
parListUpdate_func <- function(parList, doBeta = FALSE) {
    if(doBeta) {
        remInd <- which.min(em_betaVec(parList))
    } else {
        remInd <- which.min(em_gamList(parList)$gamVec)
    }
    parList <- em_JSet(parList, em_J(parList) - 1)
    gamList <- em_gamList(parList)
    gamList$gamVec <- gamList$gamVec[-remInd]
    parList <- em_zIndSet(parList, which.max(gamList$gamVec))
    gamList$gamVec <- gamList$gamVec -
        gamList$gamVec[em_zInd(parList)]
    gamList$gamDot <- sum(exp(gamList$gamVec))
    parList <- em_gamListSet(parList, gamList)
    parList <- em_betaVecSet(parList, em_betaVec(parList)[-remInd])
    parList <- em_pListSet(parList, lpMat_func(parList))
    parList <- em_gSet(parList, parGrad_func(parList))
    parList <- em_gTildeSet(parList, gTilde_func(parList))
    parList <- em_SSet(parList, matrix(0, 2 * em_J(parList), 2 * em_J(parList)))
    return(parList)
}


# takeEMStep_func is a helper function to update parList to the next iterate
takeEMStep_func <- function(parList) {
    parList <- calcStep_func(parList)

    dTheta <- em_parUpdate(parList)$parVec -
        c(em_gamList(parList)$gamVec, em_betaVec(parList))
    dg <- em_gUpdate(parList)$g - em_g(parList)$g
    parList <- em_dgTildeSet(parList, gTildeUpdate_func(parList))
    dgTilde <- em_dgTilde(parList)$gTilde - em_gTilde(parList)$gTilde

    dThetaStar <- -dgTilde + em_S(parList) %*% dg
    parList <- em_SSet(
        parList,
        em_S(parList) + c(1 + (t(dg) %*% dThetaStar) / (t(dg) %*% dTheta)) *
            (dTheta %*% t(dTheta)) / c(t(dg) %*% dTheta) -
            (dThetaStar %*% t(dTheta) + t(dThetaStar %*% t(dTheta))) /
            c(t(dg) %*% dTheta)
    )

    parList <- em_pListSet(parList, em_pListUpdate(parList))
    parList <- em_gamListSet(parList, em_parUpdate(parList)$gamList)
    parList <- em_betaVecSet(parList, em_parUpdate(parList)$betaVec)
    parList <- em_gSet(parList, em_gUpdate(parList))
    parList <- em_gTildeSet(parList, em_dgTilde(parList))

    return(parList)
}


# calcStep_func is a helper function to calculate the step direction/length
# for the modified EM iteration
calcStep_func <- function(parList) {
    # Control parameters for strong Wolfe line search
    # c1: small, strictly positive
    # c2: large, strictly less than 1, strictly greater than c1
    c1 <- 1e-3; c2 <- 0.8

    # step direction and initial step length
    parList <- em_dSet(
        parList, as.numeric(
            -em_gTilde(parList)$gTilde + em_S(parList) %*% em_g(parList)$g
        )
    )
    # Sanity check on step length
    parList <- em_alphaSet(parList, -1)
    if(any(abs(em_d(parList)) > 250)) {
        parList <- em_alphaSet(parList, -250 / max(abs(em_d(parList))))
    }

    # check initial Wolfe conditions
    parList <- em_parUpdateSet(parList, parUpdate_func(parList))
    parList <- em_lLikUpdateSet(parList, lLikUpdate_func(parList))
    w1 <- em_lLikUpdate(parList) >= em_lLik(parList) +
        c1 * em_alpha(parList) * t(em_d(parList)) %*% em_g(parList)$g
    parList <- em_pListUpdateSet(parList, lpMatUpdate_func(parList))
    parList <- em_gUpdateSet(parList, parGradUpdate_func(parList))
    w2 <- abs(t(em_d(parList)) %*% em_gUpdate(parList)$g) <=
        c2 * abs(t(em_d(parList)) %*% em_g(parList)$g)

    if(w1 & w2) {
        # Return after conditions met
        return(parList)
    } else if(t(em_g(parList)$g) %*% em_S(parList) %*% em_g(parList)$g > 0) {
        # Reset S matrix (indefinite) and take standard EM step
        parList <- resetS_func(parList)
    } else {
        parList <- alphaLine_func(parList, w1, w2, c1, c2)
    }

    if(em_alpha(parList) == 0) {
        # Reset S matrix (indefinite) and take standard EM step
        parList <- resetS_func(parList)
    }

    parList <- em_parUpdateSet(parList, parUpdate_func(parList))
    parList <- em_lLikUpdateSet(parList, lLikUpdate_func(parList))
    parList <- em_pListUpdateSet(parList, lpMatUpdate_func(parList))
    parList <- em_gUpdateSet(parList, parGradUpdate_func(parList))

    return(parList)
}


# zoom_func is a helper function to calculate the acceptable step length
# from bounds
zoom_func <- function(parList, alphaBounds) {
    # Control parameters for strong Wolfe line search
    # c1: small, strictly positive
    # c2: large, strictly less than 1, strictly greater than c1
    c1 <- 1e-3; c2 <- 0.8

    if(alphaBounds[1] != 0) {
        parList <- em_alphaSet(parList, alphaBounds[1])
        parList <- em_parUpdateSet(parList, parUpdate_func(parList))
        lLikLow <- lLikUpdate_func(parList)
    } else {
        lLikLow <- em_lLik(parList)
    }
    iter <- 0
    while(TRUE) {
        iter <- iter + 1
        if(iter == 10) {
            return(0)
        }
        parList <- em_alphaSet(parList, mean(alphaBounds))
        parList <- em_parUpdateSet(parList, parUpdate_func(parList))
        parList <- em_lLikUpdateSet(parList, lLikUpdate_func(parList))
        w1 <- em_lLikUpdate(parList) >= em_lLik(parList) +
            c1 * em_alpha(parList) * t(em_d(parList)) %*% em_g(parList)$g
        parList <- em_pListUpdateSet(parList, lpMatUpdate_func(parList))
        parList <- em_gUpdateSet(parList, parGradUpdate_func(parList))
        w2 <- abs(t(em_d(parList)) %*% em_gUpdate(parList)$g) <=
            c2 * abs(t(em_d(parList)) %*% em_g(parList)$g)
        if(!w1 | em_lLikUpdate(parList) <= lLikLow) {
            alphaBounds[2] <- em_alpha(parList)
        } else {
            if(w2) {
                return(em_alpha(parList))
            } else if(
                t(em_d(parList)) %*% em_gUpdate(parList)$g *
                diff(alphaBounds) <= 0
            ) {
                alphaBounds[2] <- alphaBounds[1]
            }
            alphaBounds[1] <- em_alpha(parList)
            lLikLow <- em_lLikUpdate(parList)
        }
    }
}


# resetS_func is a helper function to reset the EM acceleration with a strict
# EM step
resetS_func <- function(parList) {
    # Reset S matrix (indefinite) and take standard EM step
    parList <- em_SSet(parList, em_S(parList) * 0)
    parList <- em_dSet(parList, as.numeric(-em_gTilde(parList)$gTilde))
    parList <- em_parUpdateSet(parList, parUpdate_func(parList))
    parList <- em_lLikUpdateSet(parList, lLikUpdate_func(parList))
    parList <- em_pListUpdateSet(parList, lpMatUpdate_func(parList))
    parList <- em_gUpdateSet(parList, parGradUpdate_func(parList))
    parList <- em_alphaSet(parList, -1)

    return(parList)
}


# alphaLine_func is a helper function to perform line-search for the EM update
alphaLine_func <- function(parList, w1, w2, c1, c2) {
    # Line search for step length
    alphaVec <- c(0, em_alpha(parList))
    # Sanity bound on step size
    alphaMin <- -100
    iter <- 1
    while(TRUE) {
        if(!w1 | (em_lLikUpdate(parList) <= em_lLik(parList) & iter > 1)) {
            parList <- em_alphaSet(parList, zoom_func(
                parList, alphaVec[iter:(iter + 1)]
            ))
            break
        } else if (w2) {
            parList <- em_alphaSet(parList, alphaVec[(iter + 1)])
            break
        } else if (t(em_d(parList)) %*% em_gUpdate(parList)$g >= 0) {
            parList <- em_alphaSet(parList, zoom_func(
                parList, alphaVec[(iter + 1):iter]
            ))
            break
        } else {
            if(em_alpha(parList) == alphaMin) {break}
            iter <- iter + 1
            alphaVec[iter + 1] <- max(alphaVec[iter] * 2, alphaMin)
            parList <- em_alphaSet(parList, alphaVec[iter + 1])
            parList <- em_parUpdateSet(parList, parUpdate_func(parList))
            parList <- em_lLikUpdateSet(parList, lLikUpdate_func(parList))
            w1 <- em_lLikUpdate(parList) >= em_lLik(parList) +
                c1 * em_alpha(parList) *
                t(em_d(parList)) %*% em_g(parList)$g
            parList <- em_pListUpdateSet(parList, lpMatUpdate_func(parList))
            parList <- em_gUpdateSet(parList, parGradUpdate_func(parList))
            w2 <- abs(t(em_d(parList)) %*% em_gUpdate(parList)$g) <=
                c2 * abs(t(em_d(parList)) %*% em_g(parList)$g)
        }
    }

    return(parList)
}


# rqsSample_func is a helper function to perform restricted quantile sampling
# given fitted EM parameters
rqsSample_func <- function(parList, y, depth, fitSlope, prec) {
    lamPar <- pmax(exp(em_betaVec(parList)), .Machine$double.xmin * 1e1)
    bw <- tryCatch({
        suppressWarnings(bw.ucv(lamPar))
    }, error = function(e) {
        tryCatch({
            bw.nrd(lamPar)
        }, error = function(e) {
            1
        })
    })
    phi <- min(bw, 1)
    tau <- exp(em_gamList(parList)$gamVec - max(em_gamList(parList)$gamVec))
    tau <- tau / sum(tau)

    # Update model for all points
    parList <- em_ySet(parList, y)
    parList <- em_depthSet(parList, exp(depth) * fitSlope - fitSlope + 1)
    lowP <- rep(0, length(y))
    highP <- rep(0, length(y))
    nzInd <- y > 0
    for(i in seq_len(em_J(parList))) {
        lowP[nzInd] <- lowP[nzInd] + tau[i] *
            ppois(y[nzInd] - 1, lamPar[i] * em_depth(parList)[nzInd])
        highP <- highP + tau[i] * ppois(y, lamPar[i] * em_depth(parList))
    }

    # Estimate mixture CDF
    # Lower bound (minimum from rounding precision)
    lBound <- max(
        10 ^ (-prec) / 2,
        qgamma(min(lowP), min(lamPar) / phi, scale = phi)
    )
    # Upper bound (minimum to guarantee 2 steps)
    hBound <- max(
        qgamma(max(highP), max(lamPar) / phi, scale = phi), lBound + 1e-2
    )
    qSeq <- seq(lBound, hBound, min(1e-1, (hBound - lBound) / 100))
    pSeq <- rep(0, length(qSeq))
    for(i in seq_len(em_J(parList))) {
        pSeq <- pSeq + tau[i] * pgamma(qSeq, lamPar[i] / phi, scale = phi)
    }
    cdfFun <- approxfun(pSeq, qSeq, yleft = 0, yright = max(qSeq))
    normVec <- cdfFun(runif(length(y), lowP, highP))
    normVec <- round(normVec, prec)

    return(normVec)
}


# posteriorSample_func is a helper function to perform sampling from the
# estimated posterior distribution given fitted EM parameters
posteriorSample_func <- function(parList, y, depth, fitSlope, prec, emPar) {
    lamPar <- pmax(exp(em_betaVec(parList)), .Machine$double.xmin * 1e1)
    bw <- tryCatch({
        suppressWarnings(bw.ucv(lamPar))
    }, error = function(e) {
        tryCatch({
            bw.nrd(lamPar)
        }, error = function(e) {
            1
        })
    })
    phi <- min(bw, 1)

    # Update model for all points
    parList <- em_ySet(parList, y)
    parList <- em_depthSet(parList, exp(depth) * fitSlope - fitSlope + 1)
    parList <- em_pListSet(parList, lpMat_func(parList))
    pList <- em_pList(parList)
    pList$spMat <- em_pList(parList)$spMat *
        exp(em_pList(parList)$lpMax)
    parList <- em_pListSet(parList, pList)
    lamPar <- pmax(exp(em_betaVec(parList)), .Machine$double.xmin * 1e1)

    # Resample counts
    distVec <- apply(
        em_pList(parList)$spMat[-em_J(parList), , drop = FALSE], 2,
        function(pVec) {
            which.min(runif(1) > c(cumsum(pVec), 1))
        }
    )

    shapeVec <- lamPar / phi
    concVec <- emPar$conPar

    normVec <- rgamma(
        n = length(em_y(parList)),
        shape = em_y(parList) * concVec + shapeVec[distVec],
        scale = 1 / (em_depth(parList) * concVec + 1 / phi)
    )
    normVec <- round(normVec, prec)

    return(normVec)
}

# calcDepthRep is a helper function to calculate the replications in the depth
# vector. Procedure is close to that of table, but maintains the same rounding
# scheme as other functions
calcDepthRep <- function(depth) {
    depth <- sort(depth)
    dupDepth <- duplicated(depth)
    freq <- c(which(!dupDepth), length(depth) + 1)
    return(data.frame(
        depth = depth[!dupDepth],
        rep = freq[-1] - freq[-length(freq)]
    ))
}


# setPar is a helper function to initiate the parallel environment
#
#' @importFrom BiocParallel SnowParam
#' @importFrom BiocParallel MulticoreParam
#' @importFrom BiocParallel register
setPar <- function(nCores, taskList = NULL) {
    if(is.null(taskList)) {
        nTask = 0L
    } else {
        nTask <- ceiling(sqrt(length(taskList) / nCores)) * nCores
    }
    if (.Platform$OS.type == "windows") {
        prll <- SnowParam(workers = nCores, tasks = nTask, progressbar = TRUE)
        register(BPPARAM = prll, default = TRUE)
    } else {
        prll <- MulticoreParam(workers = nCores, tasks = nTask,
                    progressbar = TRUE)
        register(BPPARAM = prll, default = TRUE)
    }
    return(prll)
}


# splitGenes is a helper function to split a sparse count matrix into a list of
# gene sets for parallel computation
#
#' @importFrom Matrix Matrix
splitGenes <- function(counts, nCol = 40) {
    if(nCol >= ncol(counts)) {
        return(list(counts))
    }
    m <- floor(ncol(counts) / nCol)
    ret <- lapply(seq_len(m), function(i) {
        counts[, (nCol * i - (nCol - 1)):(nCol * i), drop = FALSE]
    })
    if(m * nCol < ncol(counts)) {
        ret[[m + 1]] <- counts[, (nCol * m + 1):ncol(counts), drop = FALSE]
    }
    return(ret)
}

#' Plot data from simulated expression
#'
#' This data is used in the vignette to demonstrate the flexibility of the
#' Dino model to smoothly estimate arbitrary latent uninmodal expression
#' distributions. These data are intended for internal use only.
#'
#' @docType data
#'
#' @usage data("unimodalDat")
#'
#' @format Object of class \code{"gtable"}.
#'
#' @keywords datasets
#'
#' @examples
#' data("unimodalDat")
#'
"unimodalDat"


#' Plot data from simulated expression
#'
#' This data is used in the vignette to demonstrate the flexibility of the
#' Dino model to smoothly estimate arbitrary latent multimodal expression
#' distributions. These data are intended for internal use only.
#'
#' @docType data
#'
#' @usage data("multimodalDat")
#'
#' @format Object of class \code{"gtable"}.
#'
#' @keywords datasets
#'
#' @examples
#' data("multimodalDat")
#'
"multimodalDat"

# S4 Class for organization of internal EM-Algorithm parameters and intermediate
# results
setClass(
    "emFitPars",
    representation(
        y = "numeric", # observed counts

        depth = "numeric", # sequencing depth / library size

        J =  "numeric", # number of mixing components

        gamList = "list", # list of current mixture weights
        # gamList$gamVec: uncentered log-transformed weights
        # gamList$gamDot: the sum for transformation to mixture weights

        zInd = "numeric", # index of largest gamList$gamVec for
        # stability centering

        betaVec = "numeric", # current estimate of (log) Poisson means

        pList = "list", # list of coefficients defining current conditional
        # mixture component membership probabilities
        # pList$spMat: row-scaled (for stability) matrix (J x length(y))
        # of conditional probabilities
        # pList$lpMax: log-scale row maxs for transformation of pList$spMat
        # back to probability space

        gTilde = "list", # list of differences between next "strict" EM iterate
        # and current parameter estimates for "accelerated" EM modification
        # gTilde$gamTilde: vector of new mixture wights minus gamList$gamVec
        # gTilde$betaTilde: vector of new Poisson means minus betaVec
        # gTilde$gTilde: c(gTilde$gamTilde, gTilde$betaTilde)

        g = "list", # list of parameter gradients for "accelerated" EM
        # g$gGam: gradient of log-likelihood in mixture weights
        # g$gBeta: gradient of log-likelihood in  Poisson means
        # g$g: c(g$gGam, g$gBeta)

        S = "matrix", # approximate inverse Hessian (2*J x 2*J) matrix for
        # "accelerated" EM updates

        lLik = "numeric", # current log-likelihood

        alpha = "numeric", # optimization step-length

        d = "numeric", # optimization step-direction

        dgTilde = "list", # list of differences between k+1 parameters and
        # k+1 "strict" EM update
        # dgTilde$gamTilde: gTilde$gamTilde at iterate k+1
        # dgTilde$betaTilde: gTilde$betaTilde at iterate k+1
        # dgTilde$gTilde: c(dgTilde$gamTilde, dgTilde$betaTilde)

        parUpdate = "list", # list of optimization parameters at next iterate
        # parUpdate$gamList: gamList at iterate k+1
        # parUpdate$betaVec: betaVec at iterate k+1
        # parUpdate$parVec: c(parUpdate$gamList$gamVec, parUpdate$betaVec)

        gUpdate = "list", # list of log-likelihood gradients at next iterate
        # gUpdate$gGam: g$gGam at iterate k + 1
        # gUpdate$gBeta: g$gBeta at iterate k + 1
        # gUpdate$g: c(gUpdate$gGam, gUpdate$gBeta)

        pListUpdate = "list", # list of conditional mixture component membership
        # probabilities at next iterate
        # pListUpdate$spMat: pList$spMat at iterate k + 1
        # pListUpdate$lpMax: pList$lpMax at iterate k + 1

        lLikUpdate = "numeric" # log-likelihood at next iterate
    )
)


# emFitPars methods
setGeneric("em_y", function(object) {
    standardGeneric("em_y")
})
setMethod(f = "em_y", signature(object = "emFitPars"), function(object) {
    return(object@y)
})
setGeneric("em_ySet", function(object, value) {
    standardGeneric("em_ySet")
})
setMethod(
    f = "em_ySet", signature(object = "emFitPars"),
    function(object, value) {
        object@y <- value
        return(object)
    }
)

setGeneric("em_depth", function(object) {
    standardGeneric("em_depth")
})
setMethod(f = "em_depth", signature(object = "emFitPars"), function(object) {
    object@depth
})
setGeneric("em_depthSet", function(object, value) {
    standardGeneric("em_depthSet")
})
setMethod(
    f = "em_depthSet", signature(object = "emFitPars"),
    function(object, value) {
        object@depth <- value
        return(object)
    }
)

setGeneric("em_J", function(object) {
    standardGeneric("em_J")
})
setMethod(f = "em_J", signature(object = "emFitPars"), function(object) {
    object@J
})
setGeneric("em_JSet", function(object, value) {
    standardGeneric("em_JSet")
})
setMethod(
    f = "em_JSet", signature(object = "emFitPars"),
    function(object, value) {
        object@J <- value
        return(object)
    }
)

setGeneric("em_gamList", function(object) {
    standardGeneric("em_gamList")
})
setMethod(f = "em_gamList", signature(object = "emFitPars"), function(object) {
    object@gamList
})
setGeneric("em_gamListSet", function(object, value) {
    standardGeneric("em_gamListSet")
})
setMethod(
    f = "em_gamListSet", signature(object = "emFitPars"),
    function(object, value) {
        object@gamList <- value
        return(object)
    }
)

setGeneric("em_zInd", function(object) {
    standardGeneric("em_zInd")
})
setMethod(f = "em_zInd", signature(object = "emFitPars"), function(object) {
    object@zInd
})
setGeneric("em_zIndSet", function(object, value) {
    standardGeneric("em_zIndSet")
})
setMethod(
    f = "em_zIndSet", signature(object = "emFitPars"), function(object, value) {
        object@zInd <- value
        return(object)
    }
)

setGeneric("em_betaVec", function(object) {
    standardGeneric("em_betaVec")
})
setMethod(f = "em_betaVec", signature(object = "emFitPars"), function(object) {
    object@betaVec
})
setGeneric("em_betaVecSet", function(object, value) {
    standardGeneric("em_betaVecSet")
})
setMethod(
    f = "em_betaVecSet", signature(object = "emFitPars"),
    function(object, value) {
        object@betaVec <- value
        return(object)
    }
)

setGeneric("em_pList", function(object) {
    standardGeneric("em_pList")
})
setMethod(f = "em_pList", signature(object = "emFitPars"), function(object) {
    object@pList
})
setGeneric("em_pListSet", function(object, value) {
    standardGeneric("em_pListSet")
})
setMethod(
    f = "em_pListSet", signature(object = "emFitPars"),
    function(object, value) {
        object@pList <- value
        return(object)
    }
)

setGeneric("em_gTilde", function(object) {
    standardGeneric("em_gTilde")
})
setMethod(f = "em_gTilde", signature(object = "emFitPars"), function(object) {
    object@gTilde
})
setGeneric("em_gTildeSet", function(object, value) {
    standardGeneric("em_gTildeSet")
})
setMethod(
    f = "em_gTildeSet", signature(object = "emFitPars"),
    function(object, value) {
        object@gTilde <- value
        return(object)
    }
)

setGeneric("em_g", function(object) {
    standardGeneric("em_g")
})
setMethod(f = "em_g", signature(object = "emFitPars"), function(object) {
    object@g
})
setGeneric("em_gSet", function(object, value) {
    standardGeneric("em_gSet")
})
setMethod(
    f = "em_gSet", signature(object = "emFitPars"),
    function(object, value) {
        object@g <- value
        return(object)
    }
)

setGeneric("em_S", function(object) {
    standardGeneric("em_S")
})
setMethod(f = "em_S", signature(object = "emFitPars"), function(object) {
    object@S
})
setGeneric("em_SSet", function(object, value) {
    standardGeneric("em_SSet")
})
setMethod(
    f = "em_SSet", signature(object = "emFitPars"),
    function(object, value) {
        object@S <- value
        return(object)
    }
)

setGeneric("em_lLik", function(object) {
    standardGeneric("em_lLik")
})
setMethod(f = "em_lLik", signature(object = "emFitPars"), function(object) {
    object@lLik
})
setGeneric("em_lLikSet", function(object, value) {
    standardGeneric("em_lLikSet")
})
setMethod(
    f = "em_lLikSet", signature(object = "emFitPars"),
    function(object, value) {
        object@lLik <- value
        return(object)
    }
)

setGeneric("em_alpha", function(object) {
    standardGeneric("em_alpha")
})
setMethod(f = "em_alpha", signature(object = "emFitPars"), function(object) {
    object@alpha
})
setGeneric("em_alphaSet", function(object, value) {
    standardGeneric("em_alphaSet")
})
setMethod(
    f = "em_alphaSet", signature(object = "emFitPars"),
    function(object, value) {
        object@alpha <- value
        return(object)
    }
)

setGeneric("em_d", function(object) {
    standardGeneric("em_d")
})
setMethod(f = "em_d", signature(object = "emFitPars"), function(object) {
    object@d
})
setGeneric("em_dSet", function(object, value) {
    standardGeneric("em_dSet")
})
setMethod(
    f = "em_dSet", signature(object = "emFitPars"),
    function(object, value) {
        object@d <- value
        return(object)
    }
)

setGeneric("em_dgTilde", function(object) {
    standardGeneric("em_dgTilde")
})
setMethod(f = "em_dgTilde", signature(object = "emFitPars"), function(object) {
    object@dgTilde
})
setGeneric("em_dgTildeSet", function(object, value) {
    standardGeneric("em_dgTildeSet")
})
setMethod(
    f = "em_dgTildeSet", signature(object = "emFitPars"),
    function(object, value) {
        object@dgTilde <- value
        return(object)
    }
)

setGeneric("em_parUpdate", function(object) {
    standardGeneric("em_parUpdate")
})
setMethod(
    f = "em_parUpdate", signature(object = "emFitPars"), function(object) {
        object@parUpdate
    }
)
setGeneric("em_parUpdateSet", function(object, value) {
    standardGeneric("em_parUpdateSet")
})
setMethod(
    f = "em_parUpdateSet", signature(object = "emFitPars"),
    function(object, value) {
        object@parUpdate <- value
        return(object)
    }
)

setGeneric("em_gUpdate", function(object) {
    standardGeneric("em_gUpdate")
})
setMethod(f = "em_gUpdate", signature(object = "emFitPars"), function(object) {
    object@gUpdate
})
setGeneric("em_gUpdateSet", function(object, value) {
    standardGeneric("em_gUpdateSet")
})
setMethod(
    f = "em_gUpdateSet", signature(object = "emFitPars"),
    function(object, value) {
        object@gUpdate <- value
        return(object)
    }
)

setGeneric("em_pListUpdate", function(object) {
    standardGeneric("em_pListUpdate")
})
setMethod(
    f = "em_pListUpdate", signature(object = "emFitPars"), function(object) {
        object@pListUpdate
    }
)
setGeneric("em_pListUpdateSet", function(object, value) {
    standardGeneric("em_pListUpdateSet")
})
setMethod(
    f = "em_pListUpdateSet", signature(object = "emFitPars"),
    function(object, value) {
        object@pListUpdate <- value
        return(object)
    }
)

setGeneric("em_lLikUpdate", function(object) {
    standardGeneric("em_lLikUpdate")
})
setMethod(
    f = "em_lLikUpdate", signature(object = "emFitPars"), function(object) {
        object@lLikUpdate
    }
)
setGeneric("em_lLikUpdateSet", function(object, value) {
    standardGeneric("em_lLikUpdateSet")
})
setMethod(
    f = "em_lLikUpdateSet", signature(object = "emFitPars"),
    function(object, value) {
        object@lLikUpdate <- value
        return(object)
    }
)

---
title: "Normalization by distributional resampling of high throughput single-cell RNA-sequencing data"
authors: "Jared Brown and Christina Kendziorski"
package: Dino
date: 05/31/2022
output: BiocStyle::html_document
vignette: >
    %\VignetteIndexEntry{Normalization by distributional resampling of high throughput single-cell RNA-sequencing data}
    %\VignetteEngine{knitr::rmarkdown}
    \usepackage[utf8]{inputenc}
---

```{r Initialize, echo=FALSE, results="hide", message=FALSE}
require(knitr)
opts_chunk$set(error=FALSE, message=FALSE, warning=FALSE)
```

# Introduction

Over the past decade, advances in single-cell RNA-sequencing (scRNA-seq) technologies have significantly increased the sensitivity and specificity with which cellular transcriptional dynamics can be analyzed. Further, parallel increases in the number cells which can be simultaneously sequenced have allowed for novel analysis pipelines including the description of transcriptional trajectories and the discovery of rare sub-populations of cells. The development of droplet-based, unique-molecular-identifier (UMI) protocols such as Drop-seq, inDrop, and the 10x Genomics Chromium platform have significantly contributed to these advances. In particular, the commercially available 10x Genomics platform has allowed the rapid and cost effective gene expression profiling of hundreds to tens of thousands of cells across many studies to date.

The use of UMIs in the 10x Genomics and related platforms has augmented these developments in sequencing technology by tagging individual mRNA transcripts with unique cell and transcript specific identifiers. In this way, biases due to transcript length and PCR amplification have been significantly reduced. However, technical variability in sequencing depth remains and, consequently, normalization to adjust for sequencing depth is required to ensure accurate downstream analyses. To address this, we introduce `Dino`, an `R` package implementing the **Dino** normalization method.

`Dino` utilizes a flexible mixture of Negative Binomials model of gene expression to reconstruct full gene-specific expression distributions which are independent of sequencing depth. By giving exact zeros positive probability, the Negative Binomial components are applicable to shallow sequencing (high proportions of zeros). Additionally, the mixture component is robust to cell heterogeneity as it accommodates multiple centers of gene expression in the distribution. By directly modeling (possibly heterogenous) gene-specific expression distributions, Dino outperforms competing approaches, especially for datasets in which the proportion of zeros is high as is typical for modern, UMI based protocols.

`Dino` does not attempt to correct for batch or other sample specific effects, and will only do so to the extent that they are correlated with sequencing depth. In situations where batch effects are expected, downstream analysis may benefit from such accommodations.

# Quick Start

## Installation

`Dino` is now available on `BioConductor` and can be easily installed from that repository by running:

```{r Install Dino BioC, eval = FALSE}
# Install Bioconductor if not present, skip otherwise
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

# Install Dino package
BiocManager::install("Dino")

# View (this) vignette from R
browseVignettes("Dino")
```

`Dino` is also available from Github, and bug fixes, patches, and updates are available there first. To install `Dino` from Github, run

```{r Install Dino Github, eval = FALSE}
devtools::install_github('JBrownBiostat/Dino', build_vignettes = TRUE)
```

*Note*, building vignettes can take a little time, so for a quicker install, consider setting `build_vignettes = FALSE`.

## All-in-one function

`Dino` (function) is an all-in-one function to normalize raw UMI count data from 10X Cell Ranger or similar protocols. Under default options, `Dino` outputs a sparse matrix of normalized expression. `SeuratFromDino` provides one-line functionality to return a Seurat object from raw UMI counts or from a previously normalized expression matrix.

```{r Quick Start, eval = FALSE}
library(Dino)

# Return a sparse matrix of normalized expression
Norm_Mat <- Dino(UMI_Mat)

# Return a Seurat object from already normalized expression
# Use normalized (doNorm = FALSE) and un-transformed (doLog = FALSE) expression
Norm_Seurat <- SeuratFromDino(Norm_Mat, doNorm = FALSE, doLog = FALSE)

# Return a Seurat object from UMI expression
# Transform normalized expression as log(x + 1) to improve
# some types of downstream analysis
Norm_Seurat <- SeuratFromDino(UMI_Mat)
```


# Detailed steps

## Read UMI data

To facilitate concrete examples, we demonstrate normalization on a small subset of sequencing data from about 3,000 peripheral blood mononuclear cells (PBMCs) published by [10X Genomics](https://support.10xgenomics.com/single-cell-gene-expression/datasets/1.1.0/pbmc3k). This dataset, named `pbmcSmall` contains 200 cells and 1,000 genes and is included with the `Dino` package.

```{r load pbmcSmall data}
set.seed(1)

# Bring pbmcSmall into R environment
library(Dino)
library(Seurat)
library(Matrix)
data("pbmcSmall")
print(dim(pbmcSmall))
```

While `Dino` was developed to normalize UMI count data, it will run on any matrix of non-negative expression data; user caution is advised if applying `Dino` to non-UMI sequencing protocols. Input formats may be sparse or dense matrices of expression with genes (features) on the rows and cells (samples) on the columns.

## Clean UMI data

While `Dino` can normalize the `pbmcSmall` dataset as it currently exists, the resulting normalized matrix, and in particular, downstream analysis are likely to be improved by cleaning the data. Of greatest use is removing genes that are expected *not* to contain useful information. This set of genes may be case dependent, but a good rule of thumb for UMI protocols is to remove genes lacking a minimum of non-zero expression prior to normalization and analysis. 

By default, `Dino` will not perform the resampling algorithm on any genes without at least 10 non-zero samples, and will rather normalize such genes by scaling with sequencing depth. To demonstrate a stricter threshold, we remove genes lacking at least 20 non-zero samples prior to normalization.

```{r clean data}
# Filter genes for a minimum of non-zero expression
pbmcSmall <- pbmcSmall[rowSums(pbmcSmall != 0) >= 20, ]
print(dim(pbmcSmall))
```

## Normalize UMI data

`Dino` contains several options to tune output. One of particular interest is `nCores` which allows for parallel computation of normalized expression. By default, `Dino` runs with two threads. Choosing `nCores = 0` will utilize all available cores, and otherwise an integer number of parallel instances can be chosen.

```{r normalize data, eval = FALSE}
# Normalize data
pbmcSmall_Norm <- Dino(pbmcSmall)
```

```{r normalize data background, echo = FALSE}
invisible(capture.output(pbmcSmall_Norm <- Dino(pbmcSmall)))
```

## Clustering with Seurat

After normalization, `Dino` makes it easy to perform data analysis. The default output is the normalized matrix in sparse format, and `Dino` additionally provides a function to transform normalized output into a `Seurat` object. We demonstrate this by running a quick clustering pipeline in `Seurat`. Much of the pipeline is modified from the tutorial at [https://satijalab.org/seurat/v3.1/pbmc3k_tutorial.html](https://satijalab.org/seurat/v3.1/pbmc3k_tutorial.html)

```{R Seurat clustering}
# Reformat normalized expression as a Seurat object
pbmcSmall_Seurat <- SeuratFromDino(pbmcSmall_Norm, doNorm = FALSE)

# Cluster pbmcSmall_Seurat
pbmcSmall_Seurat <- FindVariableFeatures(pbmcSmall_Seurat, 
                        selection.method = "mvp")
pbmcSmall_Seurat <- ScaleData(pbmcSmall_Seurat, 
                        features = rownames(pbmcSmall_Norm))
pbmcSmall_Seurat <- RunPCA(pbmcSmall_Seurat, 
                        features = VariableFeatures(object = pbmcSmall_Seurat),
                        verbose = FALSE)
pbmcSmall_Seurat <- FindNeighbors(pbmcSmall_Seurat, dims = 1:10)
pbmcSmall_Seurat <- FindClusters(pbmcSmall_Seurat, verbose = FALSE)
pbmcSmall_Seurat <- RunUMAP(pbmcSmall_Seurat, dims = 1:10)
DimPlot(pbmcSmall_Seurat, reduction = "umap")
```

## Normalizing data formatted as SingleCellExperiment

`Dino` additionally supports the normalization of datasets formatted as *SingleCellExperiment*. As with the `Seurat` pipeline, this functionality is implemented through the use of a wrapper function. We demonstrate this by quickly converting the *pbmcSmall* dataset to a *SingleCellExperiment* object and then normalizing.

```{r SinglleCellExperiment, eval = FALSE}
# Reformatting pbmcSmall as a SingleCellExperiment
library(SingleCellExperiment)
pbmc_SCE <- SingleCellExperiment(assays = list("counts" = pbmcSmall))

# Run Dino
pbmc_SCE <- Dino_SCE(pbmc_SCE)
str(normcounts(pbmc_SCE))
```

```{r SinglleCellExperiment Background, echo = F}
# Reformatting pbmcSmall as a SingleCellExperiment
library(SingleCellExperiment)
pbmc_SCE <- SingleCellExperiment(assays = list("counts" = pbmcSmall))

# Run Dino
invisible(capture.output(pbmc_SCE <- Dino_SCE(pbmc_SCE)))
str(normcounts(pbmc_SCE))
```

## Alternate sequencing depth

By default, `Dino` computes sequencing depth, which is corrected for in the normalized data, as the sum of expression for a cell (sample) across genes. This sum is then scaled such that the median depth is 1. For some datasets, however, it may be beneficial to run `Dino` on an alternately computed set of sequencing depths. *Note*: it is generally recommended that the median depth not be far from 1 as this corresponds to recomputing expression as though all cells had been sequenced at the median depth.

A simple pipeline to compute alternate sequencing depths utilizes the `Scran` method for computing normalization scale factors, and is demonstrated below.

```{r Scran depths, eval = FALSE}
library(scran)

# Compute scran size factors
scranSizes <- calculateSumFactors(pbmcSmall)

# Re-normalize data
pbmcSmall_SNorm <- Dino(pbmcSmall, nCores = 1, depth = log(scranSizes))
```

A fuller discussion of a specific use case for providing alternate sequencing depths can be viewed on the `Dino` Github page: [Issue #1](https://github.com/JBrownBiostat/Dino/issues/1)


# Method
## Model
`Dino` models observed UMI counts as a mixture of Negative Binomial random variables. The Negative Binomial distribution can, however, be decomposed into a hierarchical Gamma-Poisson distribution, so for gene $g$ and cell $j$, the `Dino` model for UMI counts is:
$$y_{gj}\sim f^{P}(\lambda_{gj}\delta_{j})\\
\lambda_{gj}\sim\sum_{K}\pi_{k}f^{G}\left(\frac{\mu_{gk}}{\theta_g},\theta_g\right)$$
where $f^{P}$ is a Poisson distribution parameterized by mean $\lambda_{gj}\delta_{j}$ and $f^{G}$ is a Gamma distribution parameterized by shape $\mu_{gk}/\theta_g$ and scale $\theta_g$. $\delta_{j}$ is the cell-specific sequencing depth, $\lambda_{gj}$ is the latent level of gene/cell-specific expression independent of depth, component probabilities $\pi_{k}$ sum to 1, the Gamma distribution is parameterized such that $\mu_{gk}$ denotes the distribution mean, and the Gamma scale paramter, $\theta_g$, is constant across mixture components.

Following model fitting for a fixed gene through an accelerated EM algorithm, `Dino` produces normalized expression values by resampling from the posterior distribution of the latent expression parameters, $\lambda_{gj}$. It can be shown that the distribution on the $\lambda_{j}$ (dropping the gene-specific subscript $g$ as calculations are repreated across genes) is a mixture of Gammas, specifically:
$$\mathbb{P}(\lambda_{j}|y_{j},\delta_j)=\sum_{K}\tau_{kj}f^{G}\left(\frac{\mu_{k}}{\theta}+\gamma y_{j},\frac{1}{\frac{1}{\theta}+\gamma\delta_j}\right)$$
where $\tau_{kj}$ denotes the conditional probability that $\lambda_{gj}$ was sampled from mixture component $k$ and $\gamma$ is a global concentration parameter. The $\tau_{kj}$ are estimated as part of the implementation of the EM algorithm in `Dino`. The adjustment from the concentration parameter can be seen as a bias in the normalized values towards a scale-factor version of normalization, since, in the limit of $\gamma$, the normalized expression for cell $j$ converges to $y_j/\delta_j$. Default values of $\gamma=15$ have proven successful.

## Mixture components $K$
Approximating the flexibility of a non-parametric method, `Dino` uses a large number of mixture components, $K$, in order to capture the full heterogeneity of expression that may exist for a given gene. The gene-specific number of components is estimated as the square root of the number of strictly positive UMI counts for a given gene. By default, $K$ is limited to be no larger than 100. In simulation, large values of $K$ are shown to successfully reconstruct both unimodal and multimodal underlying distributions. For example, when UMI counts are estimated under a single negative binomial distribution, the `Dino` fitted prior distribution (black, right panel) which is used to sample normalized expression closely matches the theoretical sampling distribution (red, right panel). Likewise, the fitted means ($\mu_k$ in the model, gray lines, left panel) span the range of the simulated data (heat map of counts, left panel), but concentrate around the theoretical mean of the sampling distribution (red, left panel).

```{r Unimodal Simulation, echo = FALSE, warning = FALSE}
library(ggplot2)
library(gridExtra)
library(ggpubr)
library(grid)

themeObj <- theme(
    title = element_text(size = 17),
    plot.subtitle = element_text(size = 15),
    axis.title = element_text(size = 14),
    axis.text = element_text(size = 12),
    strip.text = element_text(size = 14),
    legend.title = element_text(size = 14),
    legend.text = element_text(size = 12)
)

p1_func <- function(simDat, dinoLam, muVec, themeObj) {
    p1 <- ggplot(simDat, aes(x = LS, y = y)) +
        theme_classic() +
        geom_hex(aes(fill = log10(..count..))) +
        scale_fill_viridis_c() +
        labs(
            x = "log-Depth", y = "Expression (log)",
            title = "Fitted means", fill = "Count\n(log10)"
        )
    for(i in 1:length(dinoLam)) {
        p1 <- p1 +
            geom_abline(
                slope = 1, intercept = log1p(dinoLam[i]),
                col = 1, alpha = 0.5, lwd = 0.75
            )
    }
    for(i in 1:length(muVec)) {
        p1 <- p1 +
            geom_abline(
                slope = 1, intercept = log1p(muVec[i]),
                col = 2, lwd = 1.5
            )
    }
    p1 <- p1 +
        themeObj
}

p2_func <- function(plotDat, themeObj) {
    x.dens <- plotDat$x[plotDat$model == "NB"]
    y.dens <- plotDat$y[plotDat$model == "NB"]
    p2 <- ggplot(plotDat, aes(x = x, y = y, color = model)) +
        theme_classic() +
        geom_line() +
        scale_color_manual(values = 1:2) +
        labs(
            x = "Expression", y = "Density", color = "Model",
            title = "Reference vs.\nFitted distribution"
        ) +
        themeObj
    xInd <- which.max(
        y.dens < 1e-3 * max(y.dens) &
            x.dens > x.dens[which.max(y.dens)]
    )
    if(xInd > 1) {
        p2 <- p2 +
            xlim(c(
                0, 
                x.dens[which.max(
                    y.dens < 1e-3 * max(y.dens) &
                        x.dens > x.dens[which.max(y.dens)]
                )]
            ))
    }
    dinoDens <- data.frame(
        x = plotDat$x[plotDat$model == "Dino"],
        y = plotDat$y[plotDat$model == "Dino"]
    )
    if(max(y.dens) < max(dinoDens$y)) {
        p2 <- p2 +
            ylim(c(0, min(c(1.025 * max(dinoDens$y), 1.5 * max(y.dens)))))
    }
    return(p2)
}

plotSim_func <- function(plotList, themeObj) {
    p1 <- p1_func(plotList$simDat, plotList$dinoLam, plotList$muVec, themeObj)
    p2 <- p2_func(plotList$plotDat, themeObj)
    
    p <- grid.arrange(
        p1, p2, nrow = 1,
        top = textGrob(paste0("K = ", plotList$k), gp = gpar(fontsize = 17))
    )
    return(p)
}

data("unimodalDat")
plotSim_func(unimodalDat, themeObj)
```

Simulating data from a pair of Negative Binomial distributions with different means and different dispersion parameters yields similar results in the multimodal case.

```{r Multimodal Simulation, echo = FALSE, warning = FALSE}
data("multimodalDat")
plotSim_func(multimodalDat, themeObj)
```


# Session Information
```{r}
sessionInfo()
```


# Citation

If you use *Dino* in your analysis, please cite our paper:

Brown, J., Ni, Z., Mohanty, C., Bacher, R., and Kendziorski, C. (2021). "Normalization by distributional resampling of high throughput single-cell RNA-sequencing data." Bioinformatics, 37, 4123-4128. [https://academic.oup.com/bioinformatics/article/37/22/4123/6306403](https://academic.oup.com/bioinformatics/article/37/22/4123/6306403).

Other work referenced in this vignette include:

Satija, R., Farrell, J.A., Gennert, D., Schier, A.F. and Regev, A. (2015). "Spatial reconstruction of single-cell gene expression data." Nat. Biotechnol., 33, 495–502.
[https://doi.org/10.1038/nbt.3192](https://doi.org/10.1038/nbt.3192)

Amezquita, R.A., Lun, A.T.L., Becht, E., Carey, V.J., Carpp, L.N., Geistlinger, L., Marini, F., Rue-Albrecht, K., Risso, D., Soneson, C., et al. (2020). "Orchestrating single-cell analysis with Bioconductor." Nat. Methods, 17, 137–145.
[https://doi.org/10.1038/s41592-019-0654-x](https://doi.org/10.1038/s41592-019-0654-x)

Lun, A. T. L., Bach, K. and Marioni, J. C. (2016). "Pooling across cells to normalize single-cell RNA sequencing data with many zero counts." Genome Biol., 17, 1–14.
[https://doi.org/10.1186/s13059-016-0947-7](https://doi.org/10.1186/s13059-016-0947-7)

# Contact

Jared Brown: ![](JBrownEmail.jpg)

Christina Kendziorski: ![](CKendzEmail.jpg)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dino.R
\name{Dino}
\alias{Dino}
\title{Normalize scRNAseq data}
\usage{
Dino(counts, nCores = 2, prec = 3, minNZ = 10,
    nSubGene = 1e4, nSubCell = 1e4, depth = NULL, slope = NULL,
    minSlope = 1/2, maxSlope = 2, clusterSlope = TRUE,
    returnMeta = FALSE, doRQS = FALSE,
    emPar = list(maxIter = 100, tol = 0.1, conPar = 15, maxK = 100), ...)
}
\arguments{
\item{counts}{A numeric matrix object of expression counts - usually in
dgCMatrix format for memory efficiency. Column names denote cells
(samples or droplets) and row names denote genes.}

\item{nCores}{A non-negative integer scalar denoting the number of cores
which should be used. Setting nCores to 0 uses all cores as determined by
running \code{parallel::detectCores()}}

\item{prec}{A positive integer denoting the number of decimals to which to
round depth (if estimated internally via \code{depth = NULL}) and
normalized counts for computational efficiency.}

\item{minNZ}{A positive integer denoting the minimum number of non-zero
counts for a gene to be normalized by the Dino algorithm. It is
recommended to pre-filter the \emph{counts} matrix such that all genes
meet this threshold. Otherwise, genes with fewer than \emph{minNZ}
non-zeros will be scaled by depth for normalization.}

\item{nSubGene}{A positive integer denoting the number of genes to subset
for calculation of \emph{slope}.}

\item{nSubCell}{A positive integer denoting the number of samples to subset
for calculation of \emph{slope} and the EM algorithm.}

\item{depth}{A numeric vector of length equal to the columns of counts.
\emph{depth} denotes a median-centered, log-scale measure of cell-wise
sequencing depth. \code{Dino} defaults to defining depth as the
(within-cell) sum of counts across genes, followed by a log and
median-centering transformation.}

\item{slope}{A numeric scalar denoting the count-depth relationship on
the log-log scale. Typical values are close to 1 (implying a unit
increase in depth corresponds to a unit increase in expected counts on
the log-log scale), but may be higher, particularly in the case of
non-UMI protocols. \code{Dino} defaults to estimating \emph{slope}
internally.}

\item{minSlope}{A numeric scalar denoting the minimum slope. Fitted slopes
below this value will return a warning and be set to 1}

\item{maxSlope}{A numeric scalar denoting the maximum slope. Fitted slopes
above this value will return a warning and be set to 1}

\item{clusterSlope}{A logical indicating whether cells should be
pre-clustered prior to calculation of slope. Under the default where
cells are pre-clustered, cluster is used as a factor in the regression.}

\item{returnMeta}{A logical indicating whether metadata (sequencing depth
and slope) should be returned.}

\item{doRQS}{A logical indicating how normalization resampling is to be done.
By default (F), normalization is done by resampling from the full
posterior distribution. Alternately, restricted quantile sampling (RQS)
can be performed to enforce stronger preservation of expression ranks
in normalized data. Currently RQS is considered experimental.}

\item{emPar}{A list of parameters to send to the EM algorithm.
\emph{maxIter} denotes the maximum number of model updates. \emph{tol}
denotes the cutoff threshold for reductions in the log likelihood
function. \emph{conPar} denotes the concentration parameter for the
resampling. \emph{conPar = 1} implies full resampling from the fitted
distribution. As \emph{conPar} increases, the normalized expression
converges to the scale-factor normalized values. \emph{maxK} denotes the
maximum number of mixture components in the mixture model.}

\item{...}{Additional parameters to pass to \code{Scran::quickCluster}.}
}
\value{
\code{Dino} by default returns a matrix of normalized expression
    with identical dimensions as \emph{counts}. If \emph{returnMeta = TRUE},
    then \code{Dino} returns a list of normalized expression, sequencing
    depth, and slope.
}
\description{
\code{Dino} removes cell-to-cell variation in observed
counts due to the effects of sequencing depth from single-cell mRNA
sequencing experiments. \code{Dino} was particularly designed with UMI
based protocols in mind, but is applicable to non-UMI based chemistries
in the library preparation stage of sequencing.
}
\examples{
# raw data
data("pbmcSmall")
str(pbmcSmall)

# run Dino on raw expression matrix
pbmcSmall_Norm <- Dino(pbmcSmall)
str(pbmcSmall_Norm)

}
\references{
Brown, J., Ni, Z., Mohanty, C., Bacher, R. and Kendziorski, C.
    (2020) "Normalization by distributional resampling of high throughput
    single-cell RNA-sequencing data." bioRxiv.
    \href{https://doi.org/10.1101/2020.10.28.359901}{https://doi.org/10.1101/2020.10.28.359901}
}
\author{
Jared Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dino.R
\name{SeuratFromDino}
\alias{SeuratFromDino}
\title{Create Seurat object from Dino normalized data}
\usage{
SeuratFromDino(counts, doNorm = TRUE, doLog = TRUE, ...)
}
\arguments{
\item{counts}{A numeric matrix of count data, either raw (eg. UMIs) or
normalized expression.}

\item{doNorm}{A logical indicating whether to normalize the input
\emph{counts} data before exporting results to a \emph{Seurat} object.
By default, it is assumed that the contents of \emph{counts} raw
expression which should be normalized.}

\item{doLog}{A logical indicating whether normalized counts should be log
transformed with a psuedocount of 1 prior to export.}

\item{...}{Further arguments to pass to \emph{Dino}}
}
\value{
\code{SeuratFromDino} returns a Seurat object using Dino normalized
    and log transformed expression (default) for downstream analysis in the
    Seurat pipeline.

    If \emph{returnMeta = T} is passed to \emph{Dino}, then \emph{depth} and
    \emph{slope} results are stored in the \emph{Misc} slot under the
    names \emph{depth} and \emph{slope} respectively.
}
\description{
\code{SeuratFromDino} is a wrapper simplifying the export of
    \code{Dino} normalized counts to a \emph{Seurat} object for secondary
    analysis.
}
\examples{
# raw data
data("pbmcSmall")
str(pbmcSmall)

# run Dino on raw expression matrix, output Seurat object
pbmcSmall_Seurat <- SeuratFromDino(pbmcSmall)
str(pbmcSmall_Seurat)

}
\references{
Brown, J., Ni, Z., Mohanty, C., Bacher, R. and Kendziorski, C.
    (2020). "Normalization by distributional resampling of high throughput
    single-cell RNA-sequencing data." bioRxiv.
    \href{https://doi.org/10.1101/2020.10.28.359901}{https://doi.org/10.1101/2020.10.28.359901}

    Satija, R., Farrell, J.A., Gennert, D., Schier, A.F. and Regev, A.
    (2015). "Spatial reconstruction of single-cell gene expression data."
    Nat. Biotechnol., 33, 495–502.
    \href{https://doi.org/10.1038/nbt.3192}{https://doi.org/10.1038/nbt.3192}
}
\author{
Jared Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pbmcSmall.R
\docType{data}
\name{pbmcSmall}
\alias{pbmcSmall}
\title{Subset of 500 peripheral blood mononuclear cells (PBMCs) from a healthy donor}
\format{
An object of class \code{"dgCMatrix"}.
}
\source{
\href{https://support.10xgenomics.com/single-cell-gene-expression/datasets/1.1.0/pbmc3k}{3k PBMCs from a Healthy Donor}
}
\usage{
data("pbmcSmall")
}
\description{
This dataset derives from the "3k PBMCs from a Healthy Donor" public dataset
from 10X Genomics.
}
\examples{
data("pbmcSmall")
str(pbmcSmall)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dino.R
\name{Dino_SCE}
\alias{Dino_SCE}
\title{Run Dino normalization on a SingleCellExperiment dataset}
\usage{
Dino_SCE(SCE, ...)
}
\arguments{
\item{SCE}{A \emph{SingleCellExperiment} object with unnormalized count data
(eg. raw  UMIs) in the \emph{assays} slot under the name \emph{counts}.}

\item{...}{Further arguments to pass to \emph{Dino}}
}
\value{
\code{Dino_SCE} returns a \emph{SingleCellExperiment} object using
    Dino normalized expression in the \emph{assays} slot under the
    \emph{normcounts} name for downstream analysis.

    If \emph{returnMeta = T} is passed to \emph{Dino}, then \emph{depth} and
    \emph{slope} results are stored in the \emph{metadata} slot under the
    names \emph{depth} and \emph{slope} respectively.
}
\description{
\code{Dino_SCE} is a wrapper simplifying the application of the
    \emph{Dino} method to data formatted as a \emph{SingleCellExperiment}
}
\examples{
# raw data
data("pbmcSmall")
str(pbmcSmall)

# format as SingleCellExperiment
library(SingleCellExperiment)
pbmc_SCE <- SingleCellExperiment(assays = list("counts" = pbmcSmall))

# Run Dino
pbmc_SCE <- Dino_SCE(pbmc_SCE)
str(pbmc_SCE)
str(normcounts(pbmc_SCE))

}
\references{
Brown, J., Ni, Z., Mohanty, C., Bacher, R. and Kendziorski, C.
    (2020). "Normalization by distributional resampling of high throughput
    single-cell RNA-sequencing data." bioRxiv.
    \href{https://doi.org/10.1101/2020.10.28.359901}{https://doi.org/10.1101/2020.10.28.359901}

    Amezquita, R.A., Lun, A.T.L., Becht, E., Carey, V.J., Carpp, L.N.,
    Geistlinger, L., Marini, F., Rue-Albrecht, K., Risso, D., Soneson, C.,
    et al. (2020). "Orchestrating single-cell analysis with Bioconductor."
    Nat. Methods, 17, 137–145.
    \href{https://doi.org/10.1038/s41592-019-0654-x}{https://doi.org/10.1038/s41592-019-0654-x}
}
\author{
Jared Brown
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postSimPlots.R
\docType{data}
\name{multimodalDat}
\alias{multimodalDat}
\title{Plot data from simulated expression}
\format{
Object of class \code{"gtable"}.
}
\usage{
data("multimodalDat")
}
\description{
This data is used in the vignette to demonstrate the flexibility of the
Dino model to smoothly estimate arbitrary latent multimodal expression
distributions. These data are intended for internal use only.
}
\examples{
data("multimodalDat")

}
\keyword{datasets}
